# Imports do próprio módulo
from inewave._utils.leitura import Leitura
from inewave.config import MAX_ANOS_ESTUDO, MAX_ITERS
from inewave.config import NUM_VARIAVEIS_CUSTO_PMO, REES
from inewave.config import MESES, SUBMERCADOS
from .modelos.dger import DGer
from .modelos.dger import EnumImpressaoOperacao
from .modelos.dger import EnumImpressaoConvergencia
from .modelos.dger import EnumCorrecaoEnergiaDesvio
from .modelos.dger import EnumRepresentacaoSubmotorizacao
from .modelos.dger import EnumTipoReamostragem
from .modelos.dger import EnumRepresentanteAgregacao
from .modelos.dger import EnumMatrizCorrelacaoEspacial
from .modelos.dger import EnumMomentoReamostragem
from .modelos.dger import EnumInicioTesteConvergencia
from .modelos.dger import EnumSazonaliza
from .modelos.pmo import EnergiaFioLiquidaREEPMO
from .modelos.pmo import ConfiguracoesExpansaoPMO
from .modelos.pmo import RetasPerdasEngolimentoREEPMO
from .modelos.pmo import EnergiasAfluentesPMO
from .modelos.pmo import ConvergenciaPMO
from .modelos.pmo import RiscoDeficitENSPMO
from .modelos.pmo import CustoOperacaoPMO
from .modelos.pmo import PMO
# Imports de módulos externos
import os
import numpy as np  # type: ignore
from traceback import print_exc
from typing import IO, List, Tuple


class LeituraPMO(Leitura):
    """
    Realiza a leitura do arquivo pmo.dat
    existente em um diretório de saídas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de um arquivo pmo.dat, construindo
    um objeto `PMO` cujas informações são as mesmas do pmo.dat.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.

    Uma vez realizada a leitura do arquivo, as informações são guardadas
    internamente no atributo `pmo`.

    **Exemplos**

    >>> diretorio = "~/documentos/.../deck"
    >>> leitor = LeituraPMO(diretorio)
    >>> leitor.le_arquivo()
    # Ops, esqueci de pegar o objeto
    >>> pmo = leitor.pmo

    """
    str_dados_pmo = " DATA : "
    str_inicio_dger = "  DADOS GERAIS"
    str_inicio_converg = "    ITER               LIM.INF.        "
    str_inicio_risco = " ANO  RISCO   EENS  RISCO"
    str_inicio_custo_series = "                 CUSTO DE OPERACAO DAS"
    str_inicio_valor_esperado = "                 VALOR ESPERADO PARA PERI"
    str_inicio_custo_referenciado = "                     CUSTO OPERACAO R"
    str_inicio_efio_liquida = '***ENERGIA FIO D"AGUA LIQUIDA***'
    str_inicio_configs_expansao = "CONFIGURACOES POR QUALQUER MODIFICACAO"
    str_fim_dger = "CEPEL"
    str_fim_efio_liquida = "MODELO ESTRATEGICO DE GERACAO"
    str_fim_pmo = "DETECTADO NO CALCULO DA SIMULACAO FINAL"

    def __init__(self,
                 diretorio: str) -> None:
        super().__init__()
        self.diretorio = diretorio
        # PMO default, depois é substituído
        self.pmo = PMO(0,
                       0,
                       "",
                       DGer.dger_padrao(),
                       EnergiaFioLiquidaREEPMO(np.array([])),
                       ConfiguracoesExpansaoPMO(np.array([])),
                       RetasPerdasEngolimentoREEPMO(np.array([])),
                       EnergiasAfluentesPMO(),
                       EnergiasAfluentesPMO(),
                       EnergiasAfluentesPMO(),
                       {},
                       ConvergenciaPMO(np.array([])),
                       RiscoDeficitENSPMO([], np.array([])),
                       CustoOperacaoPMO(np.array([])),
                       CustoOperacaoPMO(np.array([])),
                       CustoOperacaoPMO(np.array([])))

    def le_arquivo(self) -> PMO:
        """
        Faz a leitura do arquivo `pmo.dat`.
        """
        try:
            caminho = os.path.join(self.diretorio, "pmo.dat")
            with open(caminho, "r") as arq:
                self.pmo = self._le_pmo(arq)
                return self.pmo
        except Exception:
            print_exc()
            return self.pmo

    def _le_pmo(self, arq: IO) -> PMO:
        """
        Faz a leitura do arquivo pmo.dat.
        """
        achou_dados_pmo = False
        leu_dados_pmo = False
        achou_dger = False
        achou_efio_liquida = False
        achou_configs_exp = False
        achou_convergencia = False
        achou_risco_ens = False
        achou_custo_series = False
        achou_valor_esperado = False
        achou_custo_refer = False
        linha = ""
        # Variáveis para armazenar os componentes do PMO, que será
        # construído quando acabar a leitura
        ano_pmo = 0
        mes_pmo = 0
        versao_newave = ""
        dger = DGer.dger_padrao()
        energia_liq = EnergiaFioLiquidaREEPMO(np.array([]))
        configs_exp = ConfiguracoesExpansaoPMO(np.array([]))
        retas_perdas = RetasPerdasEngolimentoREEPMO(np.array([]))
        convergencia = ConvergenciaPMO(np.array([]))
        risco_ens = RiscoDeficitENSPMO([], np.array([]))
        custo_series = CustoOperacaoPMO(np.array([]))
        valor_esp = CustoOperacaoPMO(np.array([]))
        custo_ref = CustoOperacaoPMO(np.array([]))
        while True:
            # Decide se lê uma linha nova ou usa a última lida
            linha = self._le_linha_com_backup(arq)
            if len(linha) == 0 or self._fim_arquivo(linha):
                self.pmo = PMO(ano_pmo,
                               mes_pmo,
                               versao_newave,
                               dger,
                               energia_liq,
                               configs_exp,
                               retas_perdas,
                               EnergiasAfluentesPMO(),
                               EnergiasAfluentesPMO(),
                               EnergiasAfluentesPMO(),
                               {},
                               convergencia,
                               risco_ens,
                               custo_series,
                               valor_esp,
                               custo_ref)
                break
            # Condição para iniciar uma leitura de dados
            if not achou_dados_pmo and not leu_dados_pmo:
                achou = LeituraPMO.str_dados_pmo in linha
                achou_dados_pmo = achou
            if not achou_dger:
                achou = LeituraPMO.str_inicio_dger in linha
                achou_dger = achou
            if not achou_efio_liquida:
                achou = LeituraPMO.str_inicio_efio_liquida in linha
                achou_efio_liquida = achou
            if not achou_configs_exp:
                achou = LeituraPMO.str_inicio_configs_expansao in linha
                achou_configs_exp = achou
            if not achou_convergencia:
                achou = LeituraPMO.str_inicio_converg in linha
                achou_convergencia = achou
            if not achou_risco_ens:
                achou = LeituraPMO.str_inicio_risco in linha
                achou_risco_ens = achou
            if not achou_custo_series:
                achou = LeituraPMO.str_inicio_custo_series in linha
                achou_custo_series = achou
            if not achou_valor_esperado:
                achou = LeituraPMO.str_inicio_valor_esperado in linha
                achou_valor_esperado = achou
            if not achou_custo_refer:
                achou = LeituraPMO.str_inicio_custo_referenciado in linha
                achou_custo_refer = achou
            # Quando achar, le cada parte adequadamente
            if achou_dados_pmo:
                ano_pmo, mes_pmo, versao_newave = self._le_dados_pmo(linha)
                achou_dados_pmo = False
                leu_dados_pmo = True
            if achou_dger:
                dger = self._le_dger(arq)
                achou_dger = False
            if achou_efio_liquida:
                energia_liq, retas_perdas = self._le_efio_liquida(arq)
                achou_efio_liquida = False
            if achou_configs_exp:
                configs_exp = self._le_configs_expansao(arq)
                achou_configs_exp = False
            if achou_convergencia:
                convergencia = self._le_convergencia(arq)
                achou_convergencia = False
            if achou_risco_ens:
                risco_ens = self._le_risco_ens(arq)
                achou_risco_ens = False
            if achou_custo_series:
                custo_series = self._le_tabela_custo(arq)
                achou_custo_series = False
            if achou_valor_esperado:
                valor_esp = self._le_tabela_custo(arq)
                achou_valor_esperado = False
            if achou_custo_refer:
                custo_ref = self._le_tabela_custo(arq)
                achou_custo_refer = False

        return self.pmo

    def _le_dados_pmo(self, linha: str) -> Tuple[int, int, str]:
        """
        Lê a linha com dados do PMO em questão e retorna dados
        sobre o mês e ano de estudo e a versão do NEWAVE usada.
        """
        # Procura o mês do PMO
        mes_pmo = 0
        for i, m in enumerate(MESES):
            if m in linha:
                mes_pmo = i + 1
                break
        # Quebra a linha no nome do mês
        ano_pmo = int(linha.split(MESES[mes_pmo-1])[1][3:7])
        # Encontra a versao do NW
        versao_newave = linha.split("Versao")[1].strip()
        return ano_pmo, mes_pmo, versao_newave

    def _le_dger(self, arq: IO) -> DGer:
        """
        Lê o eco dos dados gerais de entrada fornecidos ao caso.
        """
        dger = DGer.dger_padrao()
        ci = 57

        def le_parametro(linha: str):
            aux = linha[5:54]
            param = linha[ci:].strip()
            if "DURACAO DE CADA PERIODO" in aux:
                dger.duracao_estagio_op = int(param)
            elif "NUMERO DE ANOS DO HORIZONTE DE" in aux:
                dger.num_anos_estudo = int(param)
            elif "MES INICIAL DO PERIODO DE PRE" in aux:
                dger.mes_inicio_pre_estudo = int(param)
            elif "MES INICIAL DO PERIODO DE EST" in aux:
                dger.mes_inicio_estudo = int(param)
            elif "ANO INICIAL DO PERIODO DE EST" in aux:
                dger.ano_inicio_estudo = int(param)
            elif "NUMERO DE ANOS QUE PRECEDEM O HOR" in aux:
                dger.num_anos_pre_estudo = int(param)
            elif "NUMERO DE ANOS QUE SUCEDEM O HOR" in aux:
                dger.num_anos_pos_estudo = int(param)
            elif "NUMERO DE ANOS DO POS NA SIMULACAO" in aux:
                dger.num_anos_pos_sim_final = int(param)
            elif "IMPRIME DADOS DAS USINAS " in aux:
                p = param
                dger.imprime_dados_usinas = (True if p == "SIM"
                                             else False)
            elif "IMPRIME DADOS DE MERCADO" in aux:
                p = param
                dger.imprime_dados_mercados = (True if p == "SIM"
                                               else False)
            elif "IMPRIME DADOS DE ENERGIAS" in aux:
                p = param
                dger.imprime_dados_energias = (True if p == "SIM"
                                               else False)
            elif "IMPRIME PARAMETROS DO MODELO DE ENERGIA" in aux:
                p = param
                dger.imprime_dados_modelo = (True if p == "SIM"
                                             else False)
            elif "IMPRIME PARAMETROS DO RESERVATORIO EQUIV" in aux:
                p = param
                dger.imprime_dados_rees = (True if p == "SIM"
                                           else False)
            elif "NUMERO MAXIMO DE ITERACOES" in aux:
                dger.max_iteracoes = int(param)
            elif "NUMERO DE SIMULACOES" in aux:
                dger.num_sim_forward = int(param)
            elif "NUMERO DE ABERTURAS" in aux:
                dger.num_aberturas = int(param)
            elif "TOTAL DE SERIES SIMULADAS GRAVADAS" in aux:
                dger.num_series_sinteticas = int(param)
            elif "ORDEM MAXIMA DO MODELO DE ENERGI" in aux:
                dger.ordem_maxima_parp = int(param)
            elif "ANO INICIAL DO HISTORICO DE" in aux:
                dger.ano_inicial_vaz_historicas = int(param)
            elif "CALCULA VOLUME INICIAL" in aux:
                p = param
                dger.calcula_vol_inicial = (True if p == "SIM"
                                            else False)
            elif "TOLERANCIA PARA CONVERGENCIA" in aux:
                dger.tolerancia = float(param)
            elif "TAXA DE DESCONTO ANUAL (%)" in aux:
                dger.taxa_de_desconto = float(param)
            elif "IMPRIME DETALHAMENTO DA OPERACAO" in aux:
                val = (0 if param == "NAO" else 1)
                impr_op = EnumImpressaoOperacao.infere_valor(val)
                dger.impressao_operacao = impr_op
            elif "IMPRIME RESULTADOS DA CONVERGENCIA" in aux:
                val = (0 if param == "NAO" else 1)
                impr_c = EnumImpressaoConvergencia.infere_valor(val)
                dger.impressao_convergencia = impr_c
            elif "NUMERO MINIMO DE ITERACOES" in aux:
                dger.min_interacoes = int(param)
            elif "ADOCAO DE RACIONAMENTO PREVENTIVO" in aux:
                p = param
                dger.racionamento_preventivo = (True if p == "SIM"
                                                else False)
            elif "LIDA DO ARQUIVO DE MANUTENCOES" in aux:
                dger.numero_anos_manutencao_UTEs = int(param)
            elif "CONSIDERA PERDAS NA REDE DE TRANSMISSAO" in aux:
                p = param
                dger.perdas_transmissao = (True if p == "SIM"
                                           else False)
            elif "CONSIDERA OUTROS USOS DA AGUA" in aux:
                p = param
                dger.considera_desvio_dagua = (True if p == "SIM"
                                               else False)
            elif "OUTROS USOS DA AGUA VARIAVEL COM A ENERGIA" in aux:
                val = (0 if param == "NAO" else 1)
                des = EnumCorrecaoEnergiaDesvio.infere_valor(val)
                dger.correcao_energia_desvio = des
            elif "CONSIDERA CURVA GUIA DE SEGURANCA/VMINP" in aux:
                p = param
                dger.considera_curva_aversao = (True if p == "SIM"
                                                else False)
            elif "AGRUPAMENTO LIVRE" in aux:
                p = param
                dger.agrupamento_livre_interc = (True
                                                 if p == "SIM"
                                                 else False)
            elif "CONSIDERA EQUALIZACAO DE PENALIDADES DE" in aux:
                p = param
                dger.equaliza_penalidades_interc = (True
                                                    if p == "SIM"
                                                    else False)
            elif "CONSIDERA SUBMOTORIZACAO SAZONAL" in aux:
                val = (0 if param == "NAO" else
                       2 if "USINA" in param else 1)
                sm = EnumRepresentacaoSubmotorizacao.infere_valor(val)
                dger.representa_submotor = sm
            elif "CONSIDERA ORDENACAO AUTOMATICA REEs" in aux:
                p = linha[66:].strip()
                dger.ordenacao_automatica_subsist = (True
                                                     if p == "SIM"
                                                     else False)
            elif "CONSIDERA CARGAS ADICIONAIS" in aux:
                p = param
                dger.considera_cargas_adicionais = (True
                                                    if p == "SIM"
                                                    else False)
            elif "DELTA DE ZSUP" in aux:
                dger.delta_zsup = float(param[:-1])
            elif "DELTA DE ZINF" in aux:
                dger.delta_zinf = float(param[:-1])
            elif "NUMERO DE DELTAS DE ZINF CONSECUTIVOS" in aux:
                dger.deltas_consecutivos = int(param)
            elif "CONSIDERA ANTECIPACAO DE GERACAO" in aux:
                p = param
                dger.considera_despacho_gnl = (True
                                               if p == "SIM"
                                               else False)
            elif "ANTECIPACAO DE GERACAO TERMOELETRICA" in aux:
                p = param
                dger.modifica_auto_despacho_gnl = (True
                                                   if p == "SIM"
                                                   else False)
            elif "CONSIDERA GERACAO HIDRAULICA MINIMA" in aux:
                p = param
                dger.considera_ghmin = (True
                                        if p == "SIM"
                                        else False)
            elif "CONSIDERA GERENCIAMENTO EXTERNO DE PROCESSOS" in aux:
                p = param
                dger.gerenciador_externo = (True if p == "SIM"
                                            else False)
            elif "CONSIDERA COMUNICACAO EM DOIS NIVEIS" in aux:
                p = param
                dger.comunicacao_dois_niveis = (True if p == "SIM"
                                                else False)
            elif "CONSIDERA ARMAZENAMENTO LOCAL DE ARQUIVOS" in aux:
                p = param
                dger.armazenamento_local_temp = (True if p == "SIM"
                                                 else False)
            elif "CONSIDERA ALOCACAO DE ENERGIA EM MEMORIA" in aux:
                p = param
                dger.aloca_memoria_enas = (True if p == "SIM"
                                           else False)
            elif "CONSIDERA ALOCACAO DE CORTES EM MEMORIA" in aux:
                p = param
                dger.aloca_memoria_cortes = (True if p == "SIM"
                                             else False)
            elif "CONSIDERA SUPERFICIE DE AVERSAO A RISCO (SAR)" in aux:
                p = param
                dger.sar = (True if p == "SIM"
                            else False)
            elif "CONSIDERA MECANISMO DE AVERSAO AO RISCO (CVAR)" in aux:
                p = param
                dger.cvar = (True if "SIM" in p
                             else False)
            elif "DESCONSIDERA VAZAO MINIMA" in aux:
                p = param
                dger.desconsidera_vazao_minima = (True if p == "SIM"
                                                  else False)
            elif "CONSIDERA RESTRICOES ELETRICAS NO REE" in aux:
                p = param
                dger.considera_restricoes_elet = (True if p == "SIM"
                                                  else False)
            elif "CONSIDERA SELECAO DE CORTES DE BENDERS" in aux:
                p = param
                dger.selecao_cortes_benders = (True if p == "SIM"
                                               else False)
            elif "CONSIDERA JANELA DE CORTES DE BENDERS" in aux:
                p = param
                dger.janela_selecao_cortes = (True if p == "SIM"
                                              else False)
            elif "CONSIDERA REAMOSTRAGEM DE CENARIOS" in aux:
                p = param
                dger.reamostragem = (True if p == "SIM"
                                     else False)
            elif "PASSO DA REAMOSTRAGEM:" in aux:
                p = linha[33:37].strip()
                dger.passo_reamostragem = int(p)
            elif "TIPO DA REAMOSTRAGEM" in aux:
                t = (0 if "SORTEIA NOVOS RUIDOS" in param else 1)
                tipo_reamos = EnumTipoReamostragem.infere_valor(t)
                dger.tipo_reamostragem = tipo_reamos
            elif "CONSIDERA ZINF CALCULADO NO NO ZERO" in aux:
                p = param
                dger.considera_convergencia_no0 = (True
                                                   if p == "SIM"
                                                   else False)
            elif "REALIZA ACESSO A FCF PARA CONVERGENCIA" in aux:
                p = param
                dger.consulta_fcf = (True
                                     if p == "SIM"
                                     else False)
            elif "CENARIOS DE ENERGIA:" in aux:
                p = linha[36:]
                dger.impressao_ena = (True
                                      if p == "SIM"
                                      else False)
            elif "CORTES ATIVOS:" in aux:
                p = linha[36:]
                dger.impressao_cortes_ativos = (True
                                                if p == "SIM"
                                                else False)
            elif "REPRESENTANTE NO PROCESSO DE AGREGACAO" in aux:
                p = linha[45:]
                t = (1 if "CENTROIDE" in p else 0)
                ra = EnumRepresentanteAgregacao.infere_valor(t)
                dger.representante_agregacao = ra
            elif "MATRIZ DE CORRELACAO ESPACIAL CONSIDERADA" in aux:
                p = linha[48:]
                t = (1 if "MENSAL" in p else 0)
                mc = EnumMatrizCorrelacaoEspacial.infere_valor(t)
                dger.matriz_corr_espacial = mc
            elif "DESCONSIDERA CRITERIO DE CONV. ESTATISTICO" in aux:
                p = param
                dger.desconsidera_converg_estatist = (True
                                                      if p == "SIM"
                                                      else False)
            elif "MOMENTO DE REAMOSTRAGEM" in aux:
                p = linha[30:]
                t = (1 if "FORWARD" in p else 0)
                mr = EnumMomentoReamostragem.infere_valor(t)
                dger.momento_reamostragem = mr
            elif "MANTEM OS ARQUIVOS DE ENERGIAS APOS EXECUCAO" in aux:
                p = linha[51:].strip()
                dger.mantem_arquivos_ena = (True
                                            if p == "SIM"
                                            else False)
            elif "INICIO TESTE CONVERG." in aux:
                t = (1 if int(param) > 1 else 0)
                tc = EnumInicioTesteConvergencia.infere_valor(t)
                dger.inicio_teste_convergencia = tc
            elif "VOLUME MINIMO COM DATA SAZONAL NOS PERIODOS" in aux:
                p = param
                t = (1 if "SIM" in p else 0)
                sv = EnumSazonaliza.infere_valor(t)
                dger.sazonaliza_vmint = sv
            elif "VOLUME MAXIMO COM DATA SAZONAL NOS PERIODOS" in aux:
                p = param
                t = (1 if "SIM" in p else 0)
                sv = EnumSazonaliza.infere_valor(t)
                dger.sazonaliza_vmaxt = sv
            elif "VOLUME MAXIMO PENAL. SAZONAL NOS PERIODOS" in aux:
                p = param
                t = (1 if "SIM" in p else 0)
                sv = EnumSazonaliza.infere_valor(t)
                dger.sazonaliza_vminp = sv
            elif "CFUGA E CMONT SAZONAIS NOS PERIODOS" in aux:
                p = linha[53:].strip()
                t = (1 if "SIM" in p else 0)
                sv = EnumSazonaliza.infere_valor(t)
                dger.sazonaliza_cfuga_cmont = sv
            elif "CONSIDERA RESTRICOES DE LIMITES DE EMISSAO" in aux:
                p = param
                dger.restricoes_gee = (True
                                       if p == "SIM"
                                       else False)
            elif "CONSIDERA AFLUENCIA ANUAL NOS MODELOS" in aux:
                p = param
                af = (True
                      if p == "SIM"
                      else False)
                af_existente = dger.afluencia_anual_parp[1]
                dger.afluencia_anual_parp = (af, af_existente)
            elif "CONSIDERA VERIFICACAO AUTOMATICA DA ORDEM DO" in aux:
                p = linha[63:].strip()
                af = (True
                      if p == "SIM"
                      else False)
                af_existente = dger.afluencia_anual_parp[0]
                dger.afluencia_anual_parp = (af_existente, af)
            elif "CONSIDERACAO DA INCERTEZA NA GERACAO EOLICA" in aux:
                p = param
                dger.incerteza_ger_eolica = (True
                                             if p == "SIM"
                                             else False)
            elif "CONSIDERACAO DA INCERTEZA NA GERACAO SOLAR " in aux:
                p = param
                dger.incerteza_ger_solar = (True
                                            if p == "SIM"
                                            else False)
            elif "CONSIDERA RESTRICOES DE FORNECIMENTO DE GAS" in aux:
                p = param
                dger.restricoes_fornecimento_gas = (True
                                                    if p == "SIM"
                                                    else False)

        while True:
            linha = self._le_linha_com_backup(arq)
            # Confere se a parte do DGer já acabou
            if LeituraPMO.str_fim_dger in linha:
                break
            # Senão, lê mais um parâmetro
            le_parametro(linha)
        return dger

    def _le_efio_liquida(self,
                         arq: IO
                         ) -> Tuple[EnergiaFioLiquidaREEPMO,
                                    RetasPerdasEngolimentoREEPMO]:
        """
        Lê as tabelas de energia fio d'água líquida e as retas de
        perdas de energia fio d'água por REE.
        """
        # Inicia as variáveis que serão retornadas, com as tabelas
        # de energia fio d'água líquidas e as retas de perdas.
        tabela_energias: np.ndarray = np.array([])
        tabela_perdas: np.ndarray = np.array([])
        montou_tabela_energias = False
        montou_tabela_perdas = False
        achou_ree_energias = False
        achou_inicio_ree_perdas = False
        achou_ree_perdas = False
        n_rees = len(REES)
        rees_lidos = 0
        while True:
            # Confere se a leitura não acabou
            linha = self._le_linha_com_backup(arq)
            if rees_lidos == n_rees:
                return (EnergiaFioLiquidaREEPMO(tabela_energias),
                        RetasPerdasEngolimentoREEPMO(tabela_perdas))
            # Confere se já achou a tabela de energia da próxima REE
            if 'CONFIGURACAO DO PERIODO' in linha:
                achou_ree_energias = True
            # Confere se já achou a tabela de retas da próxima REE
            if "PERDAS POR ENGOLIMENTO MAXIMO" in linha:
                achou_inicio_ree_perdas = True
            if "REE:" in linha and achou_inicio_ree_perdas:
                # Salta uma linha para ter informação da REE
                achou_ree_perdas = True
            if achou_ree_energias:
                # Infere a REE em questão
                ree = REES.index(linha[6:18].strip()) + 1
                if not montou_tabela_energias:
                    # Se não leu nenhuma tabela ainda, usa a da primeira REE
                    # como exemplo para inicializar a variável completa
                    tabela_exemplo = self._le_tabela_efio(arq)
                    dims = tabela_exemplo.shape
                    tabela_energias = np.zeros((len(REES),
                                                dims[0],
                                                dims[1]))
                    tabela_energias[ree-1, :, :] = tabela_exemplo
                    montou_tabela_energias = True
                else:
                    tabela_energias[ree-1, :, :] = self._le_tabela_efio(arq)
                achou_ree_energias = False
            if achou_ree_perdas:
                ree = REES.index(linha[6:18].strip()) + 1
                if not montou_tabela_perdas:
                    # Se não leu nenhuma tabela ainda, usa a da primeira REE
                    # como exemplo para inicializar a variável completa
                    tabela_exemplo = self._le_tabela_perdas(arq)
                    dims = tabela_exemplo.shape
                    tabela_perdas = np.zeros((len(REES),
                                              dims[0],
                                              dims[1]))
                    tabela_perdas[ree-1, :, :] = tabela_exemplo
                    montou_tabela_perdas = True
                else:
                    tabela_perdas[ree-1, :, :] = self._le_tabela_perdas(arq)
                achou_ree_perdas = False
                rees_lidos += 1

    def _le_tabela_efio(self, arq: IO) -> np.ndarray:
        """
        Lê as informações de uma tabela de energias fio d'água líquidas.
        """
        tabela: List[List[float]] = []
        # Pula a linha do cabeçalho dos meses
        self._le_linha_com_backup(arq)
        n_meses = len(MESES)
        ci_ano = 5
        cf_ano = 10
        while True:
            linha = self._le_linha_com_backup(arq)
            # Confere se a tabela não acabou
            if len(linha) <= 1:
                return np.array(tabela)
            ci = 11
            nc = 7
            # Armazena as informações de uma linha
            dados_linha: List[float] = []
            dados_linha.append(float(linha[ci_ano:cf_ano]))
            for i in range(n_meses):
                cf = ci + nc
                dados_linha.append(float(linha[ci:cf]))
                ci = cf + 2
            tabela.append(dados_linha)

    def _le_configs_expansao(self, arq: IO) -> ConfiguracoesExpansaoPMO:
        """
        Lê as informações da tabela de configurações por período
        de acordo com a expansão do sistema.
        """
        n_meses = len(MESES)
        tabela = np.zeros((MAX_ANOS_ESTUDO, n_meses + 1),
                          dtype=np.int64)
        # Pula 5 linhas
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        i = 0
        while True:
            linha = self._le_linha_com_backup(arq)
            # Confere se a tabela não acabou
            if len(linha) <= 3:
                return ConfiguracoesExpansaoPMO(tabela[:i, :])
            ci = 5
            nc = 5
            # Armazena as informações de uma linha
            for j in range(n_meses + 1):
                cf = ci + nc
                tabela[i, j] = float(linha[ci:cf].strip())
                ci = cf + 1
            i += 1

    def _le_tabela_perdas(self, arq: IO) -> np.ndarray:
        """
        Lê as informações de uma tabela de retas das perdas de
        energia fio d'água.
        """
        tabela: List[List[float]] = []
        # Pula 4 linhas - entre a identificação da REE e o
        # início da tabela
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        while True:
            linha = self._le_linha_com_backup(arq)
            # Confere se a tabela não acabou
            if len(linha) <= 1:
                return np.array(tabela)
            ci = 1
            nc = 12
            # Armazena as informações de uma linha
            dados_linha: List[float] = []
            for i in range(3):
                cf = ci + nc
                dados_linha.append(float(linha[ci:cf].strip()))
                ci = cf + 1
            tabela.append(dados_linha)

    def _le_convergencia(self, arq: IO) -> ConvergenciaPMO:
        """
        Lê as linhas que formam a tabela do relatório de convergência
        da execução do NEWAVE.
        """
        # Salta duas linhas depois de identificar o início do relatório
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        # Inicia as variáveis
        tabela = np.zeros((MAX_ITERS, 6))
        i = 0
        while True:
            linha = self._le_linha_com_backup(arq)
            # Confere se já acabou
            if len(linha) < 2:
                return ConvergenciaPMO(tabela[:i, :])
            # Senão, confere se a linha não tem dados relevantes
            if not linha[4:8].strip().isnumeric():
                continue
            # Lê a linha normalmente
            it = int(linha[4:8])
            liminf = float(linha[9:31])
            zinf = float(linha[32:54])
            limsup = float(linha[55:77])
            zsup = float(linha[78:100])
            # Converte o tempo, se for informado nessa linha
            # Senão, pega da linha anterior
            str_tempo = linha[153:169]
            tempo = 0.0
            if "min" in str_tempo:
                str_horas = str_tempo.split('h')[0]
                str_min = str_tempo.split('h')[1].split('min')[0]
                str_seg = str_tempo.split('min')[1].split('s')[0]
                tempo = (float(str_seg) +
                         60 * float(str_min) +
                         3600 * float(str_horas))
            else:
                tempo = tabela[i - 1, -1]
            # Armazena na linha da tabela
            tabela[i, :] = np.array([it, liminf, zinf, limsup, zsup, tempo])
            i += 1

    def _le_risco_ens(self, arq: IO) -> RiscoDeficitENSPMO:
        """
        Lê as linhas que formam a tabela de RISCO e ENS por ano
        de estudo e subsistema.
        """
        # Salta duas linhas após a string de início ser identificada
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        # Inicializa as variáveis de interesse
        anos_estudo: List[int] = []
        linhas_tabela = MAX_ANOS_ESTUDO
        colunas_tabela = 2 * len(SUBMERCADOS)
        tabela = np.zeros((linhas_tabela, colunas_tabela))
        campos_colunas = [5, 7] * len(SUBMERCADOS)
        # Lê a tabela
        i = 0
        while True:
            linha = self._le_linha_com_backup(arq)
            # Confere se a tabela já acabou
            if len(linha) <= 1:
                # Constroi o objeto e retorna
                return RiscoDeficitENSPMO(anos_estudo, tabela[:i, :])
            anos_estudo.append(int(linha[1:5]))
            ci = 7
            for j in range(colunas_tabela):
                cf = ci + campos_colunas[j]
                tabela[i, j] = float(linha[ci:cf])
                ci = cf + 1
            i += 1

    def _le_tabela_custo(self, arq: IO) -> CustoOperacaoPMO:
        """
        Lê as linhas que formam uma das tabelas de composição
        do valor total de operação do pmo.dat.
        """
        inicio_tabela = "PARCELA           V.ESPERADO"
        iniciou = False
        # Procura o cabeçalho da tabela
        linha = ""
        while not iniciou:
            linha = self._le_linha_com_backup(arq)
            iniciou = inicio_tabela in linha
        # Pula uma linha
        self._le_linha_com_backup(arq)
        # Cria a variável para armazenar a tabela
        tabela = np.zeros((NUM_VARIAVEIS_CUSTO_PMO, 3))
        # Lê cada uma das variáveis
        abs_i = 32
        abs_f = 45
        perc_i = 46
        dp_i = 47
        dp_f = 59
        perc_i = 60
        perc_f = 67
        for i in range(NUM_VARIAVEIS_CUSTO_PMO):
            linha = self._le_linha_com_backup(arq)
            tabela[i, 0] = float(linha[abs_i:abs_f])
            tabela[i, 1] = float(linha[dp_i:dp_f])
            tabela[i, 2] = float(linha[perc_i:perc_f])
        return CustoOperacaoPMO(tabela)

    def _fim_arquivo(self, linha: str) -> bool:
        return LeituraPMO.str_fim_pmo in linha



# Rotinas de testes associadas ao arquivo vazpast.dat do NEWAVE
from inewave.newave.vazpast import LeituraVazPast, EscritaVazPast
import numpy as np  # type: ignore


leitor = LeituraVazPast("tests/_arquivos")
leitor.le_arquivo()


def test_leitura():
    assert len(leitor.vazpast.postos) > 0
    assert len(leitor.vazpast.nomes) > 0
    assert leitor.vazpast.tabela.shape[0] > 0


def test_escrita_e_leitura():
    escritor = EscritaVazPast("tests/_saidas")
    escritor.escreve_arquivo(leitor.vazpast)
    leitor2 = LeituraVazPast("tests/_saidas")
    leitor2.le_arquivo()
    assert leitor.vazpast == leitor2.vazpast


def test_eq_vazpast():
    leitor2 = LeituraVazPast("tests/_arquivos")
    leitor2.le_arquivo()
    leitor2.vazpast.tabela = np.array([])
    assert leitor.vazpast != leitor2.vazpast


def test_neq_vazpast():
    leitor2 = LeituraVazPast("tests/_arquivos")
    leitor2.le_arquivo()
    leitor2.vazpast.tabela = np.array([])
    assert leitor.vazpast != leitor2.vazpast



from enum import Enum
from typing import List, Tuple


class EnumComInferencia(Enum):
    """
    Classe base para os enums com inferência
    do enum a partir do valor dado
    """
    @classmethod
    def infere_valor(cls, v: int):
        """
        Construtor do enum a partir do valor
        """
        for e in cls:
            if e.value == v:
                return e
        return None


class EnumTipoExecucao(EnumComInferencia):
    r"""
    - COMPLETA:
    - SIMULACAO_FINAL
    """
    COMPLETA = 0
    SIMULACAO_FINAL = 1


class EnumTipoSimulacaoFinal(EnumComInferencia):
    r"""
    - NAO_SIMULA
    - SERIES_SINTETICAS
    - SERIE_HISTORICA
    - CONSISTENCIA_DADOS
    """
    NAO_SIMULA = 0
    SERIES_SINTETICAS = 1
    SERIE_HISTORICA = 2
    CONSISTENCIA_DADOS = 3


class EnumImpressaoOperacao(EnumComInferencia):
    r"""
    - SINOPSE
    - OPERACAO_DETALHADA
    """
    SINOPSE = 0
    OPERACAO_DETALHADA = 1


class EnumImpressaoConvergencia(EnumComInferencia):
    r"""
    - CONVERGENCIA_FINAL
    - CONVERGENCIA_TOTAL
    """
    CONVERGENCIA_FINAL = 0
    CONVERGENCIA_TOTAL = 1


class EnumTamanhoArquivoVazoes(EnumComInferencia):
    r"""
    - PALAVRAS_320
    - PALAVRAS_600
    """
    PALAVRAS_320 = 0
    PALAVRAS_600 = 1


class EnumTendenciaHidrologica(EnumComInferencia):
    r"""
    - NAO_LER
    - LER_POR_REE
    - LER_POR_POSTO
    """
    NAO_LER = 0
    LER_POR_REE = 1
    LER_POR_POSTO = 2


class EnumDuracaoPatamar(EnumComInferencia):
    r"""
    - SAZONAL
    - VARIAVEL_POR_ANO
    """
    SAZONAL = 0
    VARIAVEL_POR_ANO = 1


class EnumCorrecaoEnergiaDesvio(EnumComInferencia):
    r"""
    - CONSTANTE
    - VARIAVEL_COM_ARMAZENAMENTO
    """
    CONSTANTE = 0
    VARIAVEL_COM_ARMAZENAMENTO = 1


class EnumTipoGeracaoENAs(EnumComInferencia):
    r"""
    - RESIDUOS_IGUAIS_FW_BW
    - COMPENSACAO_CORR_CRUZADA_BW
    - COMPENSACAO_CORR_CRUZADA_FW_BW
    """
    RESIDUOS_IGUAIS_FW_BW = 0
    COMPENSACAO_CORR_CRUZADA_BW = 1
    COMPENSACAO_CORR_CRUZADA_FW_BW = 2


class EnumRepresentacaoSubmotorizacao(EnumComInferencia):
    r"""
    - NAO_CONSIDERA
    - SUBSISTEMA
    - USINA
    """
    NAO_CONSIDERA = 0
    SUBSISTEMA = 1
    USINA = 2


class EnumCVAR(EnumComInferencia):
    r"""
    - NAO_CONSIDERA
    - CONSTANTE_NO_TEMPO
    - VARIAVEL_NO_TEMPO
    """
    NAO_CONSIDERA = 0
    CONSTANTE_NO_TEMPO = 1
    VARIAVEL_NO_TEMPO = 2


class EnumTipoReamostragem(EnumComInferencia):
    r"""
    - RECOMBINACAO
    - PLENA
    """
    RECOMBINACAO = 0
    PLENA = 1


class EnumRepresentanteAgregacao(EnumComInferencia):
    r"""
    - MAIS_PROXIMO
    - CENTROIDE
    """
    MAIS_PROXIMO = 0
    CENTROIDE = 1


class EnumMatrizCorrelacaoEspacial(EnumComInferencia):
    r"""
    - ANUAL
    - MENSAL
    """
    ANUAL = 0
    MENSAL = 1


class EnumMomentoReamostragem(EnumComInferencia):
    r"""
    - BW_ITER_CORRESPONDENTE
    - FW_ITER_CORRESPONDENTE
    """
    BW_ITER_CORRESPONDENTE = 0
    FW_ITER_CORRESPONDENTE = 1


class EnumInicioTesteConvergencia(EnumComInferencia):
    r"""
    - PRIMEIRA_ITERACAO
    - ITERACAO_MINIMA
    """
    PRIMEIRA_ITERACAO = 0
    ITERACAO_MINIMA = 1


class EnumSazonaliza(EnumComInferencia):
    r"""
    - PRE_POS_NAO_SAZONAIS
    - PRE_POS_SAZONAIS
    """
    PRE_POS_NAO_SAZONAIS = 0
    PRE_POS_SAZONAIS = 1


class EnumRepresentacaoIncerteza(EnumComInferencia):
    r"""
    - NAO_REPRESENTA
    - HISTORICO
    - PARAMETROS_DISTRIBUICAO
    - CENARIOS
    """
    NAO_REPRESENTA = 0
    HISTORICO = 1
    PARAMETROS_DISTRIBUICAO = 2
    CENARIOS = 3


class DGer:
    """
    Classe para armazenar dados gerais de uma execução do NEWAVE.

    Os parâmetros são fornecidos ao construtor na mesma ordem em
    que são definidos no arquivo `dger.dat`.

    Apesar disso, não é recomendado criar um arquivo com o construtor
    desta classe, e sim utilizar o método `DGer.dger_padrao()`,
    que fornece um objeto com valores em todos os campos,
    que podem ser modificados.

    **Parâmetros**

    - nome_estudo : `str`
    - tipo_execucao: `EnumTipoExecucao`
    - duracao_estagio_op: `int`
    - num_anos_estudo: `int`,
    - mes_inicio_pre_estudo: `int`,
    - mes_inicio_estudo: `int`,
    - ano_inicio_estudo: `int`,
    - num_anos_pre_estudo: `int`,
    - num_anos_pos_estudo: `int`,
    - num_anos_pos_sim_final: `int`,
    - imprime_dados_usinas: `bool`,
    - imprime_dados_mercados: `bool`,
    - imprime_dados_energias: `bool`,
    - imprime_dados_modelo: `bool`,
    - imprime_dados_rees: `bool`,
    - max_iteracoes: `int`,
    - num_sim_forward: `int`,
    - num_aberturas: `int`,
    - num_series_sinteticas: `int`,
    - ordem_maxima_parp: `int`,
    - ano_inicial_vaz_historicas: `int`,
    - tamanho_arq_vaz_historicas: `EnumTamanhoArquivoVazoes`,
    - calcula_vol_inicial: `bool`,
    - vol_inicial_subsistema: `List[float]`,
    - tolerancia: `float`,
    - taxa_de_desconto: `float`,
    - tipo_simulacao_final: `EnumTipoSimulacaoFinal`,
    - impressao_operacao: `EnumImpressaoOperacao`,
    - impressao_convergencia: `EnumImpressaoConvergencia`,
    - intervalo_gravacao_relatorio: `int`,
    - min_interacoes: `int`,
    - racionamento_preventivo: `bool`,
    - numero_anos_manutencao_UTEs: `int`,
    - tendencia_hidrologica: `EnumTendenciaHidrologica`,
    - restricoes_itaipu: `bool`,
    - bidding_demanda: `bool`,
    - perdas_transmissao: `bool`,
    - el_nino: `bool`,
    - enso: `bool`,
    - duracao_patamar: `EnumDuracaoPatamar`,
    - considera_desvio_dagua: `bool`,
    - correcao_energia_desvio: `EnumCorrecaoEnergiaDesvio`,
    - considera_curva_aversao: `bool`,
    - tipo_geracao_afluencias: `EnumTipoGeracaoENAs`,
    - profundidade_risco_deficit: `Tuple[float, float]`,
    - iteracao_sim_final: `int`,
    - agrupamento_livre_interc: `bool`,
    - equaliza_penalidades_interc: `bool`,
    - representa_submotor: `EnumRepresentacaoSubmotorizacao`,
    - ordenacao_automatica_subsist: `bool`,
    - considera_cargas_adicionais: `bool`,
    - delta_zsup: `float`,
    - delta_zinf: `float`,
    - deltas_consecutivos: `int`,
    - considera_despacho_gnl: `bool`,
    - modifica_auto_despacho_gnl: `bool`,
    - considera_ghmin: `bool`,
    - sim_final_com_data: `bool`,
    - gerenciador_externo: `bool`,
    - comunicacao_dois_niveis: `bool`,
    - armazenamento_local_temp: `bool`,
    - aloca_memoria_enas: `bool`,
    - aloca_memoria_cortes: `bool`,
    - sar: `bool`,
    - cvar: `EnumCVAR`,
    - convergencia_minimo_zsup: `bool`,
    - desconsidera_vazao_minima: `bool`,
    - considera_restricoes_elet: `bool`,
    - selecao_cortes_benders: `bool`,
    - janela_selecao_cortes: `bool`,
    - reamostragem: `bool`,
    - tipo_reamostragem: `EnumTipoReamostragem`,
    - passo_reamostragem: `int`,
    - considera_convergencia_no0: `bool`,
    - consulta_fcf: `bool`,
    - impressao_ena: `bool`,
    - impressao_cortes_ativos: `bool`,
    - representante_agregacao: `EnumRepresentanteAgregacao`,
    - matriz_corr_espacial: `EnumMatrizCorrelacaoEspacial`,
    - desconsidera_converg_estatist: `bool`,
    - momento_reamostragem: `EnumMomentoReamostragem`,
    - mantem_arquivos_ena: `bool`,
    - inicio_teste_convergencia: `EnumInicioTesteConvergencia`
    - sazonaliza_vmint: `EnumSazonaliza`
    - sazonaliza_vmaxt: `EnumSazonaliza`
    - sazonaliza_vminp: `EnumSazonaliza`
    - sazonaliza_cfuga_cmont: `EnumSazonaliza`
    - restricoes_gee: `bool`
    - afluencia_anual_parp: `Tuple[bool, bool]`
    - incerteza_ger_eolica: `bool`
    - incerteza_ger_solar: `bool`
    - representacao_incerteza: `EnumRepresentacaoIncerteza`
    - restricoes_fornecimento_gas: `bool`

    """
    def __init__(self,
                 nome_estudo: str,
                 tipo_execucao: EnumTipoExecucao,
                 duracao_estagio_op: int,
                 num_anos_estudo: int,
                 mes_inicio_pre_estudo: int,
                 mes_inicio_estudo: int,
                 ano_inicio_estudo: int,
                 num_anos_pre_estudo: int,
                 num_anos_pos_estudo: int,
                 num_anos_pos_sim_final: int,
                 imprime_dados_usinas: bool,
                 imprime_dados_mercados: bool,
                 imprime_dados_energias: bool,
                 imprime_dados_modelo: bool,
                 imprime_dados_rees: bool,
                 max_iteracoes: int,
                 num_sim_forward: int,
                 num_aberturas: int,
                 num_series_sinteticas: int,
                 ordem_maxima_parp: int,
                 ano_inicial_vaz_historicas: int,
                 tamanho_arq_vaz_historicas: EnumTamanhoArquivoVazoes,
                 calcula_vol_inicial: bool,
                 vol_inicial_subsistema: List[float],
                 tolerancia: float,
                 taxa_de_desconto: float,
                 tipo_simulacao_final: EnumTipoSimulacaoFinal,
                 impressao_operacao: EnumImpressaoOperacao,
                 impressao_convergencia: EnumImpressaoConvergencia,
                 intervalo_gravacao_relatorio: int,
                 min_interacoes: int,
                 racionamento_preventivo: bool,
                 numero_anos_manutencao_UTEs: int,
                 tendencia_hidrologica: EnumTendenciaHidrologica,
                 restricoes_itaipu: bool,
                 bidding_demanda: bool,
                 perdas_transmissao: bool,
                 el_nino: bool,
                 enso: bool,
                 duracao_patamar: EnumDuracaoPatamar,
                 considera_desvio_dagua: bool,
                 correcao_energia_desvio: EnumCorrecaoEnergiaDesvio,
                 considera_curva_aversao: bool,
                 tipo_geracao_afluencias: EnumTipoGeracaoENAs,
                 profundidade_risco_deficit: Tuple[float, float],
                 iteracao_sim_final: int,
                 agrupamento_livre_interc: bool,
                 equaliza_penalidades_interc: bool,
                 representa_submotor: EnumRepresentacaoSubmotorizacao,
                 ordenacao_automatica_subsist: bool,
                 considera_cargas_adicionais: bool,
                 delta_zsup: float,
                 delta_zinf: float,
                 deltas_consecutivos: int,
                 considera_despacho_gnl: bool,
                 modifica_auto_despacho_gnl: bool,
                 considera_ghmin: bool,
                 sim_final_com_data: bool,
                 gerenciador_externo: bool,
                 comunicacao_dois_niveis: bool,
                 armazenamento_local_temp: bool,
                 aloca_memoria_enas: bool,
                 aloca_memoria_cortes: bool,
                 sar: bool,
                 cvar: EnumCVAR,
                 convergencia_minimo_zsup: bool,
                 desconsidera_vazao_minima: bool,
                 considera_restricoes_elet: bool,
                 selecao_cortes_benders: bool,
                 janela_selecao_cortes: bool,
                 reamostragem: bool,
                 tipo_reamostragem: EnumTipoReamostragem,
                 passo_reamostragem: int,
                 considera_convergencia_no0: bool,
                 consulta_fcf: bool,
                 impressao_ena: bool,
                 impressao_cortes_ativos: bool,
                 representante_agregacao: EnumRepresentanteAgregacao,
                 matriz_corr_espacial: EnumMatrizCorrelacaoEspacial,
                 desconsidera_converg_estatist: bool,
                 momento_reamostragem: EnumMomentoReamostragem,
                 mantem_arquivos_ena: bool,
                 inicio_teste_convergencia: EnumInicioTesteConvergencia,
                 sazonaliza_vmint: EnumSazonaliza,
                 sazonaliza_vmaxt: EnumSazonaliza,
                 sazonaliza_vminp: EnumSazonaliza,
                 sazonaliza_cfuga_cmont: EnumSazonaliza,
                 restricoes_gee: bool,
                 afluencia_anual_parp: Tuple[bool, bool],
                 incerteza_ger_eolica: bool,
                 incerteza_ger_solar: bool,
                 representacao_incerteza: EnumRepresentacaoIncerteza,
                 restricoes_fornecimento_gas: bool
                 ):
        self.nome_estudo = nome_estudo
        self.tipo_execucao = tipo_execucao
        self.duracao_estagio_op = duracao_estagio_op
        self.num_anos_estudo = num_anos_estudo
        self.mes_inicio_pre_estudo = mes_inicio_pre_estudo
        self.mes_inicio_estudo = mes_inicio_estudo
        self.ano_inicio_estudo = ano_inicio_estudo
        self.num_anos_pre_estudo = num_anos_pre_estudo
        self.num_anos_pos_estudo = num_anos_pos_estudo
        self.num_anos_pos_sim_final = num_anos_pos_sim_final
        self.imprime_dados_usinas = imprime_dados_usinas
        self.imprime_dados_mercados = imprime_dados_mercados
        self.imprime_dados_energias = imprime_dados_energias
        self.imprime_dados_modelo = imprime_dados_modelo
        self.imprime_dados_rees = imprime_dados_rees
        self.max_iteracoes = max_iteracoes
        self.num_sim_forward = num_sim_forward
        self.num_aberturas = num_aberturas
        self.num_series_sinteticas = num_series_sinteticas
        self.ordem_maxima_parp = ordem_maxima_parp
        self.ano_inicial_vaz_historicas = ano_inicial_vaz_historicas
        self.tamanho_arq_vaz_historicas = tamanho_arq_vaz_historicas
        self.calcula_vol_inicial = calcula_vol_inicial
        self.vol_inicial_subsistema = vol_inicial_subsistema
        self.tolerancia = tolerancia
        self.taxa_de_desconto = taxa_de_desconto
        self.tipo_simulacao_final = tipo_simulacao_final
        self.impressao_operacao = impressao_operacao
        self.impressao_convergencia = impressao_convergencia
        self.intervalo_gravacao_relatorio = intervalo_gravacao_relatorio
        self.min_interacoes = min_interacoes
        self.racionamento_preventivo = racionamento_preventivo
        self.numero_anos_manutencao_UTEs = numero_anos_manutencao_UTEs
        self.tendencia_hidrologica = tendencia_hidrologica
        self.restricoes_itaipu = restricoes_itaipu
        self.bidding_demanda = bidding_demanda
        self.perdas_transmissao = perdas_transmissao
        self.el_nino = el_nino
        self.enso = enso
        self.duracao_patamar = duracao_patamar
        self.considera_desvio_dagua = considera_desvio_dagua
        self.correcao_energia_desvio = correcao_energia_desvio
        self.considera_curva_aversao = considera_curva_aversao
        self.tipo_geracao_afluencias = tipo_geracao_afluencias
        self.profundidade_risco_deficit = profundidade_risco_deficit
        self.iteracao_sim_final = iteracao_sim_final
        self.agrupamento_livre_interc = agrupamento_livre_interc
        self.equaliza_penalidades_interc = equaliza_penalidades_interc
        self.representa_submotor = representa_submotor
        self.ordenacao_automatica_subsist = ordenacao_automatica_subsist
        self.considera_cargas_adicionais = considera_cargas_adicionais
        self.delta_zsup = delta_zsup
        self.delta_zinf = delta_zinf
        self.deltas_consecutivos = deltas_consecutivos
        self.considera_despacho_gnl = considera_despacho_gnl
        self.modifica_auto_despacho_gnl = modifica_auto_despacho_gnl
        self.considera_ghmin = considera_ghmin
        self.sim_final_com_data = sim_final_com_data
        self.gerenciador_externo = gerenciador_externo
        self.comunicacao_dois_niveis = comunicacao_dois_niveis
        self.armazenamento_local_temp = armazenamento_local_temp
        self.aloca_memoria_enas = aloca_memoria_enas
        self.aloca_memoria_cortes = aloca_memoria_cortes
        self.sar = sar
        self.cvar = cvar
        self.convergencia_minimo_zsup = convergencia_minimo_zsup
        self.desconsidera_vazao_minima = desconsidera_vazao_minima
        self.considera_restricoes_elet = considera_restricoes_elet
        self.selecao_cortes_benders = selecao_cortes_benders
        self.janela_selecao_cortes = janela_selecao_cortes
        self.reamostragem = reamostragem
        self.tipo_reamostragem = tipo_reamostragem
        self.passo_reamostragem = passo_reamostragem
        self.considera_convergencia_no0 = considera_convergencia_no0
        self.consulta_fcf = consulta_fcf
        self.impressao_ena = impressao_ena
        self.impressao_cortes_ativos = impressao_cortes_ativos
        self.representante_agregacao = representante_agregacao
        self.matriz_corr_espacial = matriz_corr_espacial
        self.desconsidera_converg_estatist = desconsidera_converg_estatist
        self.momento_reamostragem = momento_reamostragem
        self.mantem_arquivos_ena = mantem_arquivos_ena
        self.inicio_teste_convergencia = inicio_teste_convergencia
        self.sazonaliza_vmint = sazonaliza_vmint
        self.sazonaliza_vmaxt = sazonaliza_vmaxt
        self.sazonaliza_vminp = sazonaliza_vminp
        self.sazonaliza_cfuga_cmont = sazonaliza_cfuga_cmont
        self.restricoes_gee = restricoes_gee
        self.afluencia_anual_parp = afluencia_anual_parp
        self.incerteza_ger_eolica = incerteza_ger_eolica
        self.incerteza_ger_solar = incerteza_ger_solar
        self.representacao_incerteza = representacao_incerteza
        self.restricoes_fornecimento_gas = restricoes_fornecimento_gas

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre DGer avalia todos os campos,
        menos o nome do estudo.
        """
        if not isinstance(o, DGer):
            return False
        dger: DGer = o
        dif = False
        for (k, u), (_, v) in zip(self.__dict__.items(),
                                  dger.__dict__.items()):
            if k == "nome_estudo":
                continue
            if u != v:
                dif = True
        return not dif

    def eq_eco_saida(self, o: object) -> bool:
        """
        Compara o eco de saída do DGer em um `pmo.dat` com o DGer
        dado como entrada.

        ** Parâmetros **

        - o: `DGer`

        ** Sobre **

        A igualdade entre DGer no eco de saída ignora alguns
        campos.
        """
        if not isinstance(o, DGer):
            return False
        dger: DGer = o
        dif = False
        campos_ignorados = ["nome_estudo",
                            "tipo_execucao",
                            "restricoes_itaipu",
                            "tamanho_arq_vaz",
                            "vol_inicial_subsistema",
                            "tendencia_hidrologica",
                            "tipo_simulacao_final",
                            "intervalo_gravacao_relatorio",
                            "bidding_demanda",
                            "el_nino",
                            "enso",
                            "duracao_patamar",
                            "tipo_geracao_afluencias",
                            "profundidade_risco_deficit",
                            "iteracao_sim_final",
                            "sim_final_com_data",
                            "convergencia_minimo_zsup",
                            "representacao_incerteza"]
        for (k, u), (_, v) in zip(self.__dict__.items(),
                                  dger.__dict__.items()):
            if k in campos_ignorados:
                continue
            if u != v:
                dif = True
        return not dif

    @classmethod
    def dger_padrao(cls):
        """
        Retorna um DGer padrão, com valores em todos os campos.

        Devido ao grande número de parâmetros do objeto DGer, este
        método se fez necessário para possibilitar a geração de objetos
        de dados gerais em larga escala sem precisar configurar cada um
        dos parâmetros necessários no construtor.

        Desta forma, um objeto DGer pode ser instanciado e ter apenas
        os parâmetros que forem de interesse do usuário modificados,
        facilitando a realização de estudos.

        **Exemplos**

        >>> dger_com_sar = DGer.dger_padrao()
        >>> print(dger_padrao.sar)
        False
        >>> dger_com_sar.sar = True
        >>> print(dger_padrao.sar)
        True

        """
        return DGer("",
                    EnumTipoExecucao.COMPLETA,
                    1,
                    5,
                    1,
                    1,
                    2021,
                    0,
                    5,
                    0,
                    True,
                    True,
                    True,
                    True,
                    True,
                    45,
                    200,
                    20,
                    2000,
                    6,
                    1931,
                    EnumTamanhoArquivoVazoes.PALAVRAS_320,
                    True,
                    [0., 0., 0., 0.],
                    95.0,
                    12.0,
                    EnumTipoSimulacaoFinal.SERIES_SINTETICAS,
                    EnumImpressaoOperacao.OPERACAO_DETALHADA,
                    EnumImpressaoConvergencia.CONVERGENCIA_TOTAL,
                    1,
                    30,
                    False,
                    1,
                    EnumTendenciaHidrologica.LER_POR_POSTO,
                    False,
                    False,
                    False,
                    False,
                    False,
                    EnumDuracaoPatamar.VARIAVEL_POR_ANO,
                    True,
                    EnumCorrecaoEnergiaDesvio.VARIAVEL_COM_ARMAZENAMENTO,
                    True,
                    EnumTipoGeracaoENAs.RESIDUOS_IGUAIS_FW_BW,
                    (1.0, 2.5),
                    0,
                    True,
                    True,
                    EnumRepresentacaoSubmotorizacao.USINA,
                    False,
                    True,
                    10.0,
                    0.2,
                    3,
                    True,
                    True,
                    True,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    True,
                    False,
                    False,
                    True,
                    True,
                    False,
                    True,
                    EnumTipoReamostragem.PLENA,
                    0,
                    False,
                    False,
                    False,
                    False,
                    EnumRepresentanteAgregacao.CENTROIDE,
                    EnumMatrizCorrelacaoEspacial.MENSAL,
                    True,
                    EnumMomentoReamostragem.FW_ITER_CORRESPONDENTE,
                    False,
                    EnumInicioTesteConvergencia.ITERACAO_MINIMA,
                    EnumSazonaliza.PRE_POS_NAO_SAZONAIS,
                    EnumSazonaliza.PRE_POS_NAO_SAZONAIS,
                    EnumSazonaliza.PRE_POS_NAO_SAZONAIS,
                    EnumSazonaliza.PRE_POS_NAO_SAZONAIS,
                    False,
                    (True, False),
                    False,
                    False,
                    EnumRepresentacaoIncerteza.NAO_REPRESENTA,
                    False
                    )



# Imports do próprio módulo
from inewave._utils.leitura import Leitura
from inewave.config import MAX_ANOS_ESTUDO, MESES, REES, ORDEM_MAX_PARP
from inewave.newave.modelos.parp import PARp
# Imports de módulos externos
import os
import numpy as np  # type: ignore
from traceback import print_exc
from typing import IO, Dict, Tuple


class LeituraPARp(Leitura):
    """
    Realiza a leitura do arquivo `parp.dat`
    existente em um diretório de saídas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de um arquivo `parp.dat`, construindo
    um objeto `PARp` cujas informações são as mesmas do parp.dat.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.

    Uma vez realizada a leitura do arquivo, as informações são guardadas
    internamente no atributo `parp`.

    **Exemplos**

    >>> diretorio = "~/documentos/.../deck"
    >>> leitor = LeituraPARp(diretorio)
    >>> leitor.le_arquivo()
    # Ops, esqueci de pegar o objeto
    >>> parp = leitor.parp

    """
    str_inicio_serie = "SERIE  DE ENERGIAS DO REE"
    str_inicio_media = "SERIE MEDIA 12 MESES "
    str_inicio_correl_parc = "CORRELOGRAMO PARCIAL DA SERIE DE ENERGIAS"
    str_inicio_correl_cruz = "CORRELACAO CRUZADA VARIAVEL ANUAL"
    str_inicio_ordens_o = "ORDEM ORIGINAL DO MODELO AUTORREGRESSIVO"
    str_inicio_ordens_f = "ORDEM FINAL DO MODELO AUTORREGRESSIVO"
    str_inicio_correl_esp_a = "CORRELACAO ESPACIAL HISTORICA ANUAL"
    str_inicio_correl_esp_m = "CORRELACAO ESPACIAL HISTORICA MENSAL"
    str_fim_serie = "CORRELOGRAMO"
    str_fim_coefs = "SERIE DE RUIDOS"
    str_fim_parp = "////////////////////"

    def __init__(self,
                 diretorio: str) -> None:
        super().__init__()
        self.diretorio = diretorio
        # PARp default, depois é substituído
        self.parp = PARp({}, {}, {}, {}, {}, {}, {}, {}, {})

    def le_arquivo(self) -> PARp:
        """
        Faz a leitura do arquivo `parp.dat`.
        """
        try:
            caminho = os.path.join(self.diretorio, "parp.dat")
            with open(caminho, "r") as arq:
                self.parp = self._le_parp(arq)
                return self.parp
        except Exception:
            print_exc()
            return self.parp

    def _le_parp(self, arq: IO) -> PARp:
        """
        Faz a leitura do arquivo parp.dat.
        """
        achou_series = False
        achou_correl_parc = False
        achou_ordens_o_coefs = False
        achou_ordens_f_coefs = False
        achou_medias = False
        achou_correl_cruz = False
        achou_correl_esp_a = False
        achou_correl_esp_m = False
        leu_series = False
        leu_correl_parc = False
        leu_ordens_o_coefs = False
        leu_ordens_f_coefs = False
        leu_medias = False
        leu_correl_cruz = False
        leu_rees = {i: False for i in range(1, len(REES) + 1)}
        ree = 0
        ultima_cfg_lida = 0
        linha = ""
        # Variáveis para armazenar os componentes do PARp, que será
        # construído quando acabar a leitura
        n_meses = len(MESES)
        series: Dict[int, np.ndarray] = {i: np.zeros((100,
                                                      n_meses + 1,
                                                      100))
                                         for i in range(1,
                                                        len(REES) + 1)}
        medias: Dict[int, np.ndarray] = {i: np.zeros((100,
                                                      n_meses,
                                                      100))
                                         for i in range(1,
                                                        len(REES) + 1)}
        correl_p: Dict[int, np.ndarray] = {i: np.zeros((200,
                                                        n_meses))
                                           for i in range(1,
                                                          len(REES) + 1)}
        ordens_o: Dict[int, np.ndarray] = {i: np.zeros((MAX_ANOS_ESTUDO,
                                                        n_meses + 1),
                                                       dtype=np.int32)
                                           for i in range(1,
                                                          len(REES) + 1)}
        ordens_f: Dict[int, np.ndarray] = {i: np.zeros((MAX_ANOS_ESTUDO,
                                                        n_meses + 1),
                                                       dtype=np.int32)
                                           for i in range(1,
                                                          len(REES) + 1)}
        correl_c: Dict[int, np.ndarray] = {i: np.zeros((200,
                                                        n_meses + 1))
                                           for i in range(1,
                                                          len(REES) + 1)}
        n_linhas_coefs = MAX_ANOS_ESTUDO * n_meses
        coefs: Dict[int, np.ndarray] = {i: np.zeros((n_linhas_coefs,
                                                     ORDEM_MAX_PARP,
                                                     4))
                                        for i in range(1,
                                                       len(REES) + 1)}
        correl_esp_a: Dict[int, np.ndarray] = {i: np.zeros((len(REES),
                                                            len(REES)))
                                               for i in range(100)}
        correl_esp_m: Dict[int, np.ndarray] = {i: np.zeros((len(REES),
                                                            n_meses,
                                                            len(REES)))
                                               for i in range(100)}
        while True:
            # Decide se lê uma linha nova ou usa a última lida
            linha = self._le_linha_com_backup(arq)
            if self._fim_arquivo(linha):
                # Limpa as correls_esp_a e correl_esp_m com apenas as
                # cfgs lidas
                cfgs_lidas = list(range(1, ultima_cfg_lida + 1))
                for i in range(100):
                    if i not in cfgs_lidas:
                        correl_esp_a.pop(i)
                        correl_esp_m.pop(i)
                # Cria o objeto parp completo
                self.parp = PARp(ordens_o,
                                 ordens_f,
                                 coefs,
                                 series,
                                 correl_p,
                                 medias,
                                 correl_c,
                                 correl_esp_a,
                                 correl_esp_m)
                break
            # Verifica se terminou de ler tudo sobre uma REE
            # e reseta as flags
            if all([leu_ordens_o_coefs,
                    leu_ordens_f_coefs,
                    leu_series,
                    leu_correl_parc]):
                leu_ordens_o_coefs = False
                leu_ordens_f_coefs = False
                leu_series = False
                leu_correl_parc = False
                leu_medias = False
                leu_correl_cruz = False
                leu_rees[ree] = True
            # Condição para iniciar uma leitura de dados
            if not achou_series and not leu_series:
                achou = LeituraPARp.str_inicio_serie in linha
                achou_series = achou
                if achou:
                    self._configura_backup()
            if not achou_correl_parc and not leu_correl_parc:
                achou = LeituraPARp.str_inicio_correl_parc in linha
                achou_correl_parc = achou
            if not achou_ordens_o_coefs and not leu_ordens_o_coefs:
                achou = LeituraPARp.str_inicio_ordens_o in linha
                achou_ordens_o_coefs = achou
            if not achou_ordens_f_coefs and not leu_ordens_f_coefs:
                achou = LeituraPARp.str_inicio_ordens_f in linha
                achou_ordens_f_coefs = achou
            if not achou_medias and not leu_medias:
                achou = LeituraPARp.str_inicio_media in linha
                achou_medias = achou
                if achou:
                    self._configura_backup()
            if not achou_correl_cruz and not leu_correl_cruz:
                achou = LeituraPARp.str_inicio_correl_cruz in linha
                achou_correl_cruz = achou
            if not achou_correl_esp_a:
                achou = LeituraPARp.str_inicio_correl_esp_a in linha
                achou_correl_esp_a = achou
                if achou:
                    self._configura_backup()
            if not achou_correl_esp_m:
                achou = LeituraPARp.str_inicio_correl_esp_m in linha
                achou_correl_esp_m = achou
                if achou:
                    self._configura_backup()
            # Quando achar, le cada parte adequadamente
            if achou_series:
                ree = self._le_series(arq, series)
                achou_series = False
                leu_series = True
            if achou_correl_parc:
                self._le_tabela_correlograma(arq, correl_p, ree)
                achou_correl_parc = False
                leu_correl_parc = True
            if achou_ordens_o_coefs:
                self._le_tabela_ordens(arq, ree, ordens_o)
                achou_ordens_o_coefs = False
                leu_ordens_o_coefs = True
            if achou_ordens_f_coefs:
                self._le_ordens_coefs(arq,
                                      ree,
                                      ordens_f,
                                      coefs)
                achou_ordens_f_coefs = False
                leu_ordens_f_coefs = True
            if achou_medias:
                self._le_medias(arq, medias, ree)
                achou_medias = False
                leu_medias = True
            if achou_correl_cruz:
                self._le_tabela_correl_cruzada(arq, correl_c, ree)
                achou_correl_cruz = False
                leu_correl_cruz = True
            if achou_correl_esp_a:
                ultima_cfg_lida = self._le_tabela_correl_esp_a(arq,
                                                               correl_esp_a)
                achou_correl_esp_a = False
            if achou_correl_esp_m:
                self._le_tabelas_correl_esp_m(arq,
                                              correl_esp_m)
                achou_correl_esp_m = False

        return self.parp

    def _le_series(self,
                   arq: IO,
                   series: Dict[int, np.ndarray]) -> int:
        """
        Lê as tabelas de séries de energia para uma REE.
        """
        ree = 0
        cfg = 0
        achou = False
        while True:
            # Confere se a leitura não acabou
            linha = self._le_linha_com_backup(arq)
            if LeituraPARp.str_fim_serie in linha:
                series[ree] = series[ree][:, :, :cfg]
                break
            # Senão, lê mais uma tabela
            if not achou:
                achou = LeituraPARp.str_inicio_serie in linha
                if achou:
                    self._configura_backup()
                continue
            if achou:
                self._configura_backup()
                ree, cfg = self._le_tabela_serie(arq, series)
                achou = False
        return ree

    def _le_medias(self,
                   arq: IO,
                   medias: Dict[int, np.ndarray],
                   ree: int):
        """
        Lê as tabelas de séries de energia para uma REE.
        """
        ind_ano = 0
        achou = False
        acabou = False
        while True:
            # Confere se a leitura não acabou
            linha = self._le_linha_com_backup(arq)
            if LeituraPARp.str_inicio_correl_cruz in linha or acabou:
                self._configura_backup()
                medias[ree] = medias[ree][:, :, :ind_ano]
                break
            # Senão, lê mais uma tabela
            if not achou:
                achou = LeituraPARp.str_inicio_media in linha
                if achou:
                    self._configura_backup()
                continue
            if achou:
                self._configura_backup()
                if self._le_tabela_media(arq, medias, ree, ind_ano):
                    ind_ano += 1
                else:
                    acabou = True
                achou = False

    def _le_tabela_serie(self,
                         arq: IO,
                         series: Dict[int, np.ndarray]
                         ) -> Tuple[int, int]:
        """
        """
        # Lê a linha do backup para descobrir a REE
        linha = self._le_linha_com_backup(arq)
        str_ree = linha.split("REE")[1][:16].strip()
        ree = REES.index(str_ree) + 1
        # Descobre a configuração
        cfg = int(linha.split("CONFIGURACAO No.")[1][:-2].strip())
        # Salta 4 linhas
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        # Lê a tabela
        i = 0
        n_meses = len(MESES)
        while True:
            # Verifica se a tabela já acabou
            linha = self._le_linha_com_backup(arq)
            if len(linha) < 3 or "MEDIA AMOSTRAL DAS ENERGIAS" in linha:
                series[ree] = series[ree][:i, :, :]
                return ree, cfg
            # Senão, lê mais uma linha
            # Ano
            series[ree][i, 0, cfg-1] = int(linha[:4])
            # Energias de cada mês
            ci = 5
            nc = 9
            for j in range(n_meses):
                cf = ci + nc
                series[ree][i, j+1, cfg-1] = float(linha[ci:cf])
                ci = cf + 2
            i += 1

    def _le_tabela_correlograma(self,
                                arq: IO,
                                correl: Dict[int, np.ndarray],
                                ree: int):
        """
        """
        # Salta 2 linhas
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        # Lê a tabela
        i = 0
        n_meses = len(MESES)
        while True:
            # Verifica se a tabela já acabou
            linha = self._le_linha_com_backup(arq)
            if len(linha) < 3 or not linha[5:9].isnumeric():
                correl[ree] = correl[ree][:i, :]
                break
            # Senão, lê mais uma linha
            # Ano
            correl[ree][i, 0] = int(linha[5:9])
            # Correlação de cada mês
            ci = 11
            nc = 8
            for j in range(n_meses - 1):
                cf = ci + nc
                correl[ree][i, j+1] = float(linha[ci:cf])
                ci = cf + 2
            i += 1

    def _le_tabela_correl_cruzada(self,
                                  arq: IO,
                                  correl: Dict[int, np.ndarray],
                                  ree: int):
        """
        """
        # Salta 4 linhas
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        # Lê a tabela
        i = 0
        n_meses = len(MESES)
        while True:
            # Verifica se a tabela já acabou
            linha = self._le_linha_com_backup(arq)
            if len(linha) < 3 or not linha[5:9].isnumeric():
                correl[ree] = correl[ree][:i, :]
                break
            # Senão, lê mais uma linha
            # Ano
            correl[ree][i, 0] = int(linha[5:9])
            # Correlação de cada mês
            ci = 11
            nc = 8
            for j in range(n_meses):
                cf = ci + nc
                correl[ree][i, j+1] = float(linha[ci:cf])
                ci = cf + 2
            i += 1

    def _le_tabela_media(self,
                         arq: IO,
                         medias: Dict[int, np.ndarray],
                         ree: int,
                         ind_ano: int,
                         ) -> bool:
        """
        """
        # Lê a linha do backup para descobrir o ano
        linha = self._le_linha_com_backup(arq)
        str_ano = linha.split("ANO:")[1].strip()
        # Se é um ano de pós estudo, ignora
        if not str_ano.isnumeric():
            return False
        # Salta 5 linhas
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        # Lê a tabela
        i = 0
        n_meses = len(MESES)
        while True:
            # Verifica se a tabela já acabou
            linha = self._le_linha_com_backup(arq)
            if len(linha) < 3:
                medias[ree] = medias[ree][:i, :, :]
                return True
            # Senão, lê mais uma linha
            ci = 5
            nc = 9
            for j in range(n_meses):
                cf = ci + nc
                medias[ree][i, j, ind_ano] = float(linha[ci:cf])
                ci = cf + 2
            i += 1

    def _le_ordens_coefs(self,
                         arq: IO,
                         ree: int,
                         ordens: Dict[int, np.ndarray],
                         coefs: Dict[int, np.ndarray]):
        """
        Lê as informações das ordens e dos coeficientes do PAR(p).
        """
        n_anos = self._le_tabela_ordens(arq, ree, ordens)
        self._le_coeficientes(arq, ree, coefs, n_anos)

    def _le_tabela_ordens(self,
                          arq: IO,
                          ree: int,
                          ordens: Dict[int, np.ndarray]
                          ) -> int:
        """
        Retorna o número de anos de estudo válidos (sem POS).
        """
        # Salta 3 linhas
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        i = 0
        n_meses = len(MESES)
        while True:
            linha = self._le_linha_com_backup(arq)
            # Confere se a tabela já acabou
            if len(linha) < 3:
                ordens[ree] = ordens[ree][:i, :]
                break
            # Extrai o ano
            str_ano = linha[32:36]
            if not str_ano.isnumeric():
                continue
            ordens[ree][i, 0] = int(str_ano)
            # As ordens de cada mês
            ci = 38
            nc = 3
            for j in range(1, n_meses + 1):
                cf = ci + nc
                ordens[ree][i, j] = int(linha[ci:cf])
                ci = cf + 2
            i += 1
        return i

    def _le_coeficientes(self,
                         arq: IO,
                         ree: int,
                         coefs: Dict[int, np.ndarray],
                         n_anos: int):
        i = 0
        n_meses = len(MESES)
        while True:
            linha = self._le_linha_com_backup(arq)
            # Verifica se a leitura dos coeficientes já não terminou
            if LeituraPARp.str_fim_coefs in linha or i == n_anos * n_meses:
                coefs[ree] = coefs[ree][:i, :, :]
                break
            # Senão, lê mais uma tabela de coeficientes
            self._le_tabela_coeficientes(arq, ree, coefs, i)
            i += 1

    def _le_tabela_coeficientes(self,
                                arq: IO,
                                ree: int,
                                coefs: Dict[int, np.ndarray],
                                i: int):
        achou = False
        ordem = 0
        lin = 0
        while True:
            linha = self._le_linha_com_backup(arq)
            if not achou:
                # Procura pela ordem do modelo
                if "COEFICIENTES DA EQUACAO" in linha:
                    achou = True
                    ordem = int(linha.split("AR(")[1][:-2])
                continue
            if lin == 4 or len(linha) < 2:
                break
            if achou:
                ci = 0
                nc = 9
                if lin < 2:
                    for j in range(ordem):
                        cf = ci + nc
                        coefs[ree][i, j, lin] = float(linha[ci:cf])
                        ci = cf + 2
                else:
                    coefs[ree][i, 0, lin] = float(linha[ci:cf])
                lin += 1

    def _le_tabela_correl_esp_a(self,
                                arq: IO,
                                correls: Dict[int, np.ndarray]
                                ) -> int:
        """
        """
        # Lê a linha do backup para descobrir a configuração
        self._configura_backup()
        linha = self._le_linha_com_backup(arq)
        str_cfg = linha.split("No:")[1].strip()
        # Descobre a configuração
        cfg = int(str_cfg)
        # Salta 1 linhas
        self._le_linha_com_backup(arq)
        linha = self._le_linha_com_backup(arq)
        # Descobre a ordem das REEs nas colunas
        str_rees = [s for s in linha.split(" ") if len(s) > 1]
        ordem_rees = [REES.index(s) + 1 for s in str_rees]
        # Lê a tabela
        i = 0
        n_meses = len(MESES)
        while True:
            # Verifica se a tabela já acabou
            linha = self._le_linha_com_backup(arq)
            if len(linha) < 3:
                return cfg
            # Senão, lê mais uma linha
            ci = 18
            nc = 7
            for j in range(n_meses):
                cf = ci + nc
                correls[cfg][ordem_rees[i]-1,
                             ordem_rees[j]-1] = float(linha[ci:cf])
                ci = cf + 6
            i += 1

    def _le_tabelas_correl_esp_m(self,
                                 arq: IO,
                                 correls: Dict[int, np.ndarray]):
        """
        """
        # Lê a linha do backup para descobrir a configuração
        self._configura_backup()
        linha = self._le_linha_com_backup(arq)
        str_cfg = linha.split("No:")[1].strip()
        # Descobre a configuração
        cfg = int(str_cfg)
        # Salta 1 linha
        self._le_linha_com_backup(arq)
        # Lê a tabela
        i = 0
        while True:
            # Verifica se a tabela já acabou
            linha = self._le_linha_com_backup(arq)
            if (LeituraPARp.str_inicio_correl_esp_a in linha or
                    LeituraPARp.str_fim_parp in linha):
                self._configura_backup()
                return cfg
            # Senão, procura e lê mais uma tabela
            if "MES" in linha:
                self._configura_backup()
                self._le_tabela_correl_esp_m(arq,
                                             correls,
                                             cfg)
                i += 1

    def _le_tabela_correl_esp_m(self,
                                arq: IO,
                                correls: Dict[int, np.ndarray],
                                cfg: int):
        """
        """
        linha = self._le_linha_com_backup(arq)
        # Descobre a ordem das REEs nas colunas
        str_rees = [s for s in linha[18:].split(" ") if len(s) > 1]
        ordem_rees = [REES.index(s) for s in str_rees]
        # Lê a tabela
        i = 0
        while True:
            # Verifica se a tabela já acabou
            linha = self._le_linha_com_backup(arq)
            if len(linha) < 3:
                break
            # Senão, lê mais uma linha
            # Identifica a REE da linha
            ree_linha = REES.index(linha[:12].strip())
            ci = 18
            nc = 7
            for j in ordem_rees:
                cf = ci + nc
                correls[cfg][ree_linha,
                             i,
                             j] = float(linha[ci:cf])
                ci = cf + 6
            i += 1

    def _fim_arquivo(self, linha: str) -> bool:
        return LeituraPARp.str_fim_parp in linha



SUBMERCADOS = ["SUDESTE", "SUL", "NORDESTE", "NORTE"]
REES = ["SUDESTE",
        "SUL",
        "NORDESTE",
        "NORTE",
        "ITAIPU",
        "MADEIRA",
        "TPIRES",
        "BMONTE",
        "MAN-AP",
        "PARANA",
        "IGUACU",
        "PRNPANEMA"]
NUM_CENARIOS = 2000
MAX_ANOS_ESTUDO = 30
MAX_ITERS = 200
NUM_PATAMARES = 3
NUM_VARIAVEIS_MEDIAS = 52
NUM_VARIAVEIS_CUSTO_PMO = 16
ORDEM_MAX_PARP = 6
MESES = ["JANEIRO",
         "FEVEREIRO",
         "MARCO",
         "ABRIL",
         "MAIO",
         "JUNHO",
         "JULHO",
         "AGOSTO",
         "SETEMBRO",
         "OUTUBRO",
         "NOVEMBRO",
         "DEZEMBRO"]



# Rotinas de testes associadas ao arquivo dger.dat do NEWAVE
from inewave.newave.modelos.dger import DGer
from inewave.newave.dger import LeituraDGer, EscritaDGer


leitor = LeituraDGer("tests/_arquivos")
leitor.le_arquivo()


def test_leitura():
    assert len(leitor.dger.nome_estudo) > 0


def test_ano_inicio_estudo():
    assert leitor.dger.ano_inicio_estudo == 1995


def test_tolerancia():
    assert leitor.dger.tolerancia == 95.0


def test_risco_deficit():
    assert leitor.dger.profundidade_risco_deficit == (1.0, 2.5)


def test_reamostragem():
    assert leitor.dger.reamostragem


def test_dger_padrao():
    dger_padrao = DGer.dger_padrao()
    escritor = EscritaDGer("tests/_saidas")
    escritor.escreve_arquivo(dger_padrao)
    leitor_padrao = LeituraDGer("tests/_saidas")
    leitor_padrao.le_arquivo()
    assert leitor_padrao.dger == dger_padrao


def test_escrita_e_leitura():
    escritor = EscritaDGer("tests/_saidas")
    escritor.escreve_arquivo(leitor.dger)
    leitor2 = LeituraDGer("tests/_saidas")
    leitor2.le_arquivo()
    assert leitor.dger == leitor2.dger


def test_eq_dger():
    leitor2 = LeituraDGer("tests/_arquivos")
    leitor2.le_arquivo()
    assert leitor2.dger == leitor.dger


def test_neq_dger():
    leitor2 = LeituraDGer("tests/_arquivos")
    leitor2.le_arquivo()
    leitor2.dger.delta_zsup = -10.0
    assert leitor2.dger != leitor.dger
    assert leitor2.dger is not None



from abc import abstractmethod
from typing import IO, List
import os


class Leitura:
    """
    Classe com utilidades gerais para leitura de arquivos
    do NEWAVE.
    """
    def __init__(self):
        self.usa_backup = False
        self.linha_backup = ""
        self.diretorio = ""

    def _le_linha_com_backup(self, arq: IO) -> str:
        """
        Faz uma leitura de linha de um arquivo, mas com a opção de usar
        um backup de leitura anterior sinalizado anteriormente.
        """
        linha = ""
        if self.usa_backup:
            self.usa_backup = False
            linha = self.linha_backup
        else:
            linha = arq.readline()
            self.linha_backup = linha
        return linha

    def _configura_backup(self):
        """
        Prepara a próxima leitura para ser uma feita a partir de um
        backup armazenado.
        """
        self.usa_backup = True

    def _lista_arquivos_por_chave(self, chave: str) -> List[str]:
        """
        Retorna a lista de caminhos completos para os arquivos em um
        diretório, desde que tenham uma certa chave no nome.
        """
        return [f for f in os.listdir(self.diretorio) if chave in f]

    @abstractmethod
    def _fim_arquivo(self, linha: str) -> bool:
        """
        Método que deve ser implementado para cada arquivo, com o
        conteúdo da linha que indica o fim do próprio, para impedir loops
        de leitura eterna.
        """
        pass



# Rotinas de testes associadas ao arquivo earmfpm00x.out do NWLISTOP
from inewave.nwlistop.earmfpm00 import LeituraEarmfpm00
from inewave.config import MESES, NUM_CENARIOS
import numpy as np  # type: ignore


sub_teste = "SUDESTE"
leitor = LeituraEarmfpm00("tests/_arquivos")
leitor.le_arquivos()
anos_estudo_teste = [1995, 1996, 1997, 1998, 1999]


def test_leitura():
    assert sub_teste in leitor.earmfpms


def test_eq_earmfpm00():
    leitor2 = LeituraEarmfpm00("tests/_arquivos")
    leitor2.le_arquivos()
    assert leitor.earmfpms == leitor2.earmfpms


def test_neq_earmfpm00():
    leitor2 = LeituraEarmfpm00("tests/_arquivos")
    leitor2.le_arquivos()
    leitor2.earmfpms[sub_teste].ano_pmo = 2077
    assert leitor.earmfpms != leitor2.earmfpms


def test_extrai_dados_execucao():
    earmfpm = leitor.earmfpms[sub_teste]
    assert earmfpm.mes_pmo == 5
    assert earmfpm.ano_pmo == 1995
    assert earmfpm.submercado == sub_teste


def test_anos_estudo():
    earmfpm = leitor.earmfpms[sub_teste]
    anos_lidos = list(earmfpm.energias_armazenadas.keys())
    assert anos_estudo_teste == anos_lidos


def test_earmfpm_por_ano():
    earmfpm = leitor.earmfpms[sub_teste]
    por_ano = earmfpm.energias_por_ano
    assert anos_estudo_teste == list(por_ano.keys())
    mes = earmfpm.mes_pmo
    ano = earmfpm.ano_pmo
    # Confere se os valores médios de energias nos meses anteriores
    # ao estudo são nulos, e todos os posteriores são > 0
    assert np.all(por_ano[ano][:, :mes-1] == 0.0)
    assert np.all(por_ano[ano][:, mes-1:] > 0.0)
    for a in list(por_ano.keys())[1:]:
        assert np.all(por_ano[a] > 0.0)


def test_earmfpm_por_ano_e_mes():
    earmfpm = leitor.earmfpms[sub_teste]
    por_ano_e_mes = earmfpm.energias_por_ano_e_mes
    assert anos_estudo_teste == list(por_ano_e_mes.keys())
    mes = earmfpm.mes_pmo
    ano = earmfpm.ano_pmo
    # Confere se os valores médios de energias nos meses anteriores
    # ao estudo são nulos, e todos os posteriores são > 0
    for m in range(1, mes):
        assert np.all(por_ano_e_mes[ano][m] == 0.0)
    for m in range(mes, len(MESES) + 1):
        assert np.all(por_ano_e_mes[ano][m] > 0.0)
    for a in list(por_ano_e_mes.keys())[1:]:
        for m in range(mes, len(MESES) + 1):
            assert np.all(por_ano_e_mes[a][m] > 0.0)


def test_earmfpm_por_cenario():
    earmfpm = leitor.earmfpms[sub_teste]
    por_ano_cenario = earmfpm.energias_por_ano_e_cenario
    mes_pmo = earmfpm.mes_pmo
    # Confere se os valores lidos são nulos até o mês anterior
    # ao PMO, e positivos do mês do PMO em diante.
    anos = anos_estudo_teste
    n_meses = len(MESES)
    for c in range(NUM_CENARIOS):
        assert por_ano_cenario[anos[0]][c].shape == (n_meses,)
        assert np.all(por_ano_cenario[anos[0]][c][:mes_pmo-1] == 0)
        assert np.all(por_ano_cenario[anos[0]][c][mes_pmo-1:] > 0)
    for a in anos[1:]:
        for c in range(NUM_CENARIOS):
            assert por_ano_cenario[a][c].shape == (n_meses,)
            assert np.all(por_ano_cenario[a][c] > 0)



import numpy as np  # type: ignore


class DSVAgua:
    """
    Armazena os dados de entrada do NEWAVE referentes aos
    desvios de água por usina.

    Esta classe lida com informações de entrada fornecidas ao NEWAVE e
    que podem ser modificadas através do arquivo `dsvagua.dat`.

    **Parâmetros**

    - tabela: `np.ndarray`

    """
    def __init__(self,
                 tabela: np.ndarray):
        self.tabela = tabela

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre DSVAgua avalia toda a tabela.
        """
        if not isinstance(o, DSVAgua):
            return False
        dsv: DSVAgua = o
        return np.array_equal(self.tabela, dsv.tabela)



from typing import Dict


class UHEConfhd:
    """
    Armazena dados de uma UHE conforme configurada no
    arquivo `confhd.dat`.

    **Parâmetros**

    - numero: `int`
    - nome: `str`
    - posto: `int`
    - jusante: `int`
    - ree: `int`
    - volume_inicial: `float`
    - existente: `bool`
    - modificada: `bool`
    - numero: `int`
    - inicio_historico: `int`
    - fim_historico: `int`

    """
    __slots__ = ["numero",
                 "nome",
                 "posto",
                 "jusante",
                 "ree",
                 "volume_inicial",
                 "existente",
                 "modificada",
                 "inicio_historico",
                 "fim_historico"]

    def __init__(self,
                 numero: int,
                 nome: str,
                 posto: int,
                 jusante: int,
                 ree: int,
                 volume_inicial: float,
                 existente: bool,
                 modificada: bool,
                 inicio_historico: int,
                 fim_historico: int):
        self.numero = numero
        self.nome = nome
        self.posto = posto
        self.jusante = jusante
        self.ree = ree
        self.volume_inicial = volume_inicial
        self.existente = existente
        self.modificada = modificada
        self.inicio_historico = inicio_historico
        self.fim_historico = fim_historico

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre UHEConfhd avalia todos os campos.
        """
        if not isinstance(o, UHEConfhd):
            return False
        uhe: UHEConfhd = o
        dif = False
        for s1, s2 in zip(self.__slots__,
                          uhe.__slots__):
            if s1 != s2 or getattr(self, s1) != getattr(uhe, s2):
                dif = True
                break
        return not dif


class Confhd:
    """
    Armazena os dados de entrada do NEWAVE referentes às
    configurações das usinas hidrelétricas.

    Esta classe lida com informações de entrada fornecidas ao NEWAVE e
    que podem ser modificadas através do arquivo `modif.dat`.

    **Parâmetros**

    - usinas: `Dict[int, UHEConfhd]`

    """
    def __init__(self,
                 usinas: Dict[int, UHEConfhd]):
        self.usinas = usinas

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre Confhd avalia todas as usinas.
        """
        if not isinstance(o, Confhd):
            return False
        confhd: Confhd = o
        dif = False
        for (n1, uhe1), (n2, uhe2) in zip(self.usinas.items(),
                                          confhd.usinas.items()):
            if n1 != n2 or uhe1 != uhe2:
                dif = True
                break
        return not dif



from typing import List
import numpy as np  # type: ignore


class VazPast:
    """
    Armazena os dados de entrada do NEWAVE referentes às
    vazões anteriores ao período de planejamento.

    Esta classe lida com informações de entrada fornecidas ao NEWAVE e
    que são usadas juntos das contidas no arquivo `vazoes.dat`.

    **Parâmetros**
    - mes_planejamento: `int`
    - ano_planejamento: `int`
    - postos: `List[int]`
    - nomes: `List[str]`
    - tabela: `np.ndarray`

    """
    def __init__(self,
                 mes_planejamento: int,
                 ano_planejamento: int,
                 postos: List[int],
                 nomes: List[str],
                 tabela: np.ndarray):
        self.mes_planejamento = mes_planejamento
        self.ano_planejamento = ano_planejamento
        self.postos = postos
        self.nomes = nomes
        self.tabela = tabela

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre VazPast avalia todas as usinas.
        """
        if not isinstance(o, VazPast):
            return False
        vaz: VazPast = o
        eq_postos = self.postos == vaz.postos
        eq_nomes = self.nomes == vaz.nomes
        eq_tabela = np.array_equal(self.tabela, vaz.tabela)
        return eq_postos and eq_nomes and eq_tabela



from .dger import DGer

from typing import Dict, List
import numpy as np  # type: ignore


class EnergiasAfluentesPMO:
    """
    Armazena as informações de energias
    afluentes anteriores ao estudo contidas no arquivo `pmo.dat`.

    **Parâmetros**
    """
    def __init__(self):
        # TODO - estruturar as propriedades
        pass

    def __eq__(self, o: object):
        return True


class EnergiaFioLiquidaREEPMO:
    """
    Armazena as informações de energias a fio d'água líquidas
    para cada REE existentes no arquivo `pmo.dat`,
    quando feita a simulação completa.

    **Parâmetros**

    - tabela: `np.ndarray`

    """

    def __init__(self,
                 tabela: np.ndarray):
        self.tabela = tabela

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre EnergiaFioLiquidaREEPMO avalia todos os campos.
        """
        if not isinstance(o, EnergiaFioLiquidaREEPMO):
            return False
        e: EnergiaFioLiquidaREEPMO = o
        return np.array_equal(self.tabela, e.tabela)


class ConfiguracoesExpansaoPMO:
    """
    Armazena as informações das configurações válidas para cada
    mês do estudo contidas no arquivo `pmo.dat`.

    **Parâmetros**

    - tabela: `np.ndarray`
    """
    def __init__(self,
                 tabela: np.ndarray):
        self.tabela = tabela

    def __eq__(self, o: object):
        """
        A igualdade entre ConfiguracoesExpansaoPMO avalia todos os campos.
        """
        if not isinstance(o, ConfiguracoesExpansaoPMO):
            return False
        e: ConfiguracoesExpansaoPMO = o
        return np.array_equal(self.tabela, e.tabela)

    @property
    def configs_por_ano(self) -> Dict[int, np.ndarray]:
        """
        Configurações ativas para serem consideradas em cada
        mês de estudo, organizada por ano.

        **Retorna**
        `Dict[int, np.ndarray]`
        """
        dict_configs: Dict[int, np.ndarray] = {}
        for lin in range(self.tabela.shape[0]):
            dict_configs[self.tabela[lin, 0]] = self.tabela[lin, 1:]

        return dict_configs


class RetasPerdasEngolimentoREEPMO:
    """
    Armazena as retas que modelam as perdas por engolimento
    máximo para cada REE, existentes no arquivo `pmo.dat`, quando
    feita a simulação completa.

    **Parâmetros**

    - tabela: `np.ndarray`

    """
    def __init__(self,
                 tabela: np.ndarray):
        self.tabela = tabela

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre RetasPerdasEngolimentoREEPMO
        avalia todos os campos.
        """
        if not isinstance(o, RetasPerdasEngolimentoREEPMO):
            return False
        e: RetasPerdasEngolimentoREEPMO = o
        return np.array_equal(self.tabela, e.tabela)

    def funcao_perdas(self,
                      ree: int,
                      energia_bruta: float) -> float:
        """
        Valor da função de perdas composta pelas retas
        para um determinado valor de energia bruta fornecido.

        **Parâmetros**

        - ree: `int`
        - energia_bruta: `float`

        **Retorna**
        `float`

        """
        n_lin = self.tabela.shape[0]
        # Calcula o valor das três retas na ordenada dada e
        # retorna o maior deles.
        energia = energia_bruta * np.ones((n_lin, 1))
        perdas: np.ndarray = np.multiply(self.tabela[ree-1, :, 1],
                                         energia) + self.tabela[ree-1, :, 2]
        return float(np.max(perdas))


class DemandaLiquidaEnergiaPMO:
    """
    Armazena as informações de demandas
    líquidas de energia contidas no arquivo `pmo.dat`.

    **Parâmetros**
    """
    def __init__(self):
        # TODO - estruturar as propriedades
        pass


class ConvergenciaPMO:
    """
    Armazena as informações do relatório de convergência do NEWAVE
    existentes no arquivo `pmo.dat`.

    **Parâmetros**

    - anos_estudo: `List[int]`
    - tabela: `np.ndarray`
    """
    def __init__(self,
                 tabela: np.ndarray):
        self.tabela = tabela

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre ConvergenciaPMO
        avalia todos os campos.
        """
        if not isinstance(o, ConvergenciaPMO):
            return False
        e: ConvergenciaPMO = o
        return np.array_equal(self.tabela, e.tabela)

    @property
    def zinf(self) -> Dict[int, List[float]]:
        """
        Limite inferior da função objetivo na PDDE (Zinf).

        **Retorna**

        `Dict[int, List[float]]`

        **Sobre**

        O número de chaves do dicionário é igual ao de iterações.
        """
        zinfs: Dict[int, List[float]] = {}
        for lin in range(self.tabela.shape[0]):
            it = int(self.tabela[lin, 0])
            if it not in zinfs:
                zinfs[it] = []
            zinfs[it].append(self.tabela[lin, 2])
        return zinfs

    @property
    def zsup(self) -> Dict[int, List[float]]:
        """
        Limite superior da função objetivo na PDDE (Zsup).

        **Retorna**

        `Dict[int, List[float]]`

        **Sobre**

        O número de chaves do dicionário é igual ao de iterações.
        """
        zsups: Dict[int, List[float]] = {}
        for lin in range(self.tabela.shape[0]):
            it = int(self.tabela[lin, 0])
            if it not in zsups:
                zsups[it] = []
            zsups[it].append(self.tabela[lin, 4])
        return zsups

    @property
    def tempos_execucao(self) -> List[float]:
        """
        Tempo de execução de cada iteração em segundos.

        **Retorna**

        `List[float]`

        **Sobre**

        O número de elementos da lista é igual ao de iterações.
        """
        tempos: List[float] = []
        ultima_it = -1
        for lin in range(self.tabela.shape[0]):
            it = self.tabela[lin, 0]
            if it == ultima_it:
                continue
            tempos.append(self.tabela[lin, -1])
            ultima_it = it
        return tempos


class RiscoDeficitENSPMO:
    """
    Armazena as informações risco de déficit e valores esperados
    de energia não supridacontidas no arquivo `pmo.dat`.

    **Parâmetros**

    - anos_estudo: `List[int]`
    - tabela: `np.ndarray`
    """
    def __init__(self,
                 anos_estudo: List[int],
                 tabela: np.ndarray):
        # São impressos no pmo.dat os subsistemas em
        # uma ordem pré-determinada
        self.subsistemas = ["SUDESTE",
                            "SUL",
                            "NORDESTE",
                            "NORTE"]
        self.anos_estudo = anos_estudo
        self.tabela = tabela

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre RiscoDeficitENSPMO
        avalia todos os campos.
        """
        if not isinstance(o, RiscoDeficitENSPMO):
            return False
        e: RiscoDeficitENSPMO = o
        eq_anos = all([a == b
                       for (a, b) in zip(self.anos_estudo,
                                         e.anos_estudo)])
        eq_tabela = np.array_equal(self.tabela, e.tabela)
        return eq_anos and eq_tabela

    @property
    def riscos_por_subsistema_e_ano(self) -> Dict[str,
                                                  Dict[int,
                                                       float]]:
        """
        Riscos de déficit agrupados por
        subsistema e ano de estudo.

        **Retorna**

        `Dict[str, Dict[int, float]]`

        **Sobre**

        O acesso é feito com
        [sub][ano] e o valor fornecido é em percentual, de 0 a 100.
        """
        riscos: Dict[str, Dict[int, float]] = {}
        for j, sub in enumerate(self.subsistemas):
            if sub not in riscos:
                riscos[sub] = {}
            for i, ano in enumerate(self.anos_estudo):
                riscos[sub][ano] = self.tabela[i, 2*j]
        return riscos

    @property
    def ens_por_subsistema_e_ano(self) -> Dict[str,
                                               Dict[int,
                                                    float]]:
        """
        Energias não supridas agrupadas por
        subsistema e ano de estudo.

        **Retorna**

        `Dict[str, Dict[int, float]]`

        **Sobre**

        O acesso é feito com
        [sub][ano] e o valor fornecido é em MWmes.
        """
        energias: Dict[str, Dict[int, float]] = {}
        for j, sub in enumerate(self.subsistemas):
            if sub not in energias:
                energias[sub] = {}
            for i, ano in enumerate(self.anos_estudo):
                energias[sub][ano] = self.tabela[i, 2*j + 1]
        return energias


class CustoOperacaoPMO:
    """
    Armazena as informações do relatório
    de custo de operação, disponível no arquivo `pmo.dat`.

    Esta classe armazena uma das tabelas existentes ao final do arquivo
    `pmo.dat`, contendo os custos, os desvios-padrão e a participação
    em percentual de cada componente de custo.

    A tabela de custos é armazenada através de uma array
    em `NumPy`, para otimizar cálculos futuros e espaço ocupado
    em memória. A tabela interna é transformada em dicionários
    e outras estruturas de dados mais palpáveis através das propriedades
    da própria classe.

    **Parâmetros**

    - custos: `np.ndarray`

    """
    def __init__(self,
                 custos: np.ndarray):
        self.custos = custos
        # TODO - fazer as @property de cada uma individualmente

        # self.geracao_termica = geracao_termica
        # self.deficit = deficit
        # self.vertimento = vertimento
        # self.excesso_energia = excesso_energia
        # self.violacao_car = violacao_car
        # self.violacao_sar = violacao_sar
        # self.violacao_outro_usos = violacao_outro_usos
        # self.violacao_evmin = violacao_evmin
        # self.violacao_vzmin = violacao_vzmin
        # self.intercambio = intercambio
        # self.violacao_intercambio_min = violacao_intercambio_min
        # self.vertimento_fio_nao_turbin = vertimento_fio_nao_turbin
        # self.violacao_ghmin = violacao_ghmin
        # self.violacao_ghmin_usina = violacao_ghmin_usina
        # self.violacao_retirada = violacao_retirada
        # self.violacao_emissao_gee = violacao_emissao_gee

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre CustoOperacaoPMO
        avalia todos os campos.
        """
        if not isinstance(o, CustoOperacaoPMO):
            return False
        e: CustoOperacaoPMO = o
        return np.array_equal(self.custos, e.custos)

    @property
    def geracao_termica(self) -> float:
        """
        Parcela do custo de operação devido à geração térmica,
        em MMR$.

        **Retorna**

        `float`
        """
        return float(self.custos[0, 0])

    @property
    def deficit(self) -> float:
        """
        Parcela do custo de operação devido ao déficit, em MMR$.

        **Retorna**

        `float`
        """
        return float(self.custos[1, 0])

    @property
    def custos_seguranca(self) -> float:
        """
        Parcela do custo de operação devido a violações de segurança,
        CAR e SAR, em MMR$.

        **Retorna**

        `float`
        """
        return np.sum(self.custos[4:6, 0])

    @property
    def custos_hidricos(self) -> float:
        """
        Parcela do custo de operação devido a violações de restrições
        hídricas, EVmin, VZmin, GHmin, GHmin Usina e outros usos
        da água, em MMR$.

        **Retorna**

        `float`
        """
        return (np.sum(self.custos[6:9, 0]) +
                np.sum(self.custos[12:14, 0]))

    @property
    def outros_custos(self) -> float:
        """
        Parcela do custo de operação devido a violações e penalidades
        de intercâmbio, intercâmbio mínimo, vertimento, vertimento
        fio d'água não turbinável, excesso de energia, emissão de GEE
        e retirada, em MMR$.

        **Retorna**

        `float`
        """
        return (np.sum(self.custos[2:4, 0]) +
                np.sum(self.custos[9:12, 0]) +
                np.sum(self.custos[14:, 0]))

    @property
    def custo_total(self) -> float:
        """
        Custo total de operação, em MMR$.

        **Retorna**

        `float`
        """
        return np.sum(self.custos[:, 0])


class PMO:
    """
    Armazena os dados de entrada do NEWAVE referentes ao
    acompanhamento do programa.

    Esta classe lida com as informações de entrada fornecidas ao
    NEWAVE e reproduzidas no `pmo.dat`, bem como as saídas finais
    da execução: custos de operação, energias, déficit, etc.

    Em versões futuras, esta classe pode passar a ler os dados
    de execução intermediárias do programa.

    **Parâmetros**

    - ano_pmo: `int`
    - mes_pmo: `int`
    - versao_newave: `str`
    - dados_gerais: `DGer`
    - energias_passadas_politica: `EnergiasAfluentesPMO`
    - energias_passadas_primeira_conf: `EnergiasAfluentesPMO`
    - energias_passadas_canal_fuga: `EnergiasAfluentesPMO`
    - configuracoes_expansao: `ConfiguracoesExpansaoPMO`
    - demanda_liquida_energia : `Dict[str, DemandaLiquidaEnergiaPMO]`
    - convergencia: `ConvergenciaPMO`
    - risco_ens: `RiscoDeficitENSPMO`
    - custo_series_simuladas: `CustoOperacaoPMO`
    - valor_esperado_periodo: `CustoOperacaoPMO`
    - custo_referenciado: `CustoOperacaoPMO`

    """
    def __init__(self,
                 ano_pmo: int,
                 mes_pmo: int,
                 versao_newave: str,
                 dados_gerais: DGer,
                 energia_fio_liquida: EnergiaFioLiquidaREEPMO,
                 configuracoes_expansao: ConfiguracoesExpansaoPMO,
                 retas_perdas_engolimento: RetasPerdasEngolimentoREEPMO,
                 energias_passadas_politica: EnergiasAfluentesPMO,
                 energias_passadas_primeira_conf: EnergiasAfluentesPMO,
                 energias_passadas_canal_fuga: EnergiasAfluentesPMO,
                 demanda_liquida_energia: Dict[str,
                                               DemandaLiquidaEnergiaPMO],
                 convergencia: ConvergenciaPMO,
                 risco_ens: RiscoDeficitENSPMO,
                 custo_series_simuladas: CustoOperacaoPMO,
                 valor_esperado_periodo: CustoOperacaoPMO,
                 custo_referenciado: CustoOperacaoPMO):
        self.ano_pmo = ano_pmo
        self.mes_pmo = mes_pmo
        self.versao_newave = versao_newave
        self.dados_gerais = dados_gerais
        self.energia_fio_liquida = energia_fio_liquida
        self.configuracoes_expansao = configuracoes_expansao
        self.retas_perdas_engolimento = retas_perdas_engolimento
        self.energias_passadas_politica = energias_passadas_politica
        self.energias_passadas_primeira_conf = energias_passadas_primeira_conf
        self.energias_passadas_canal_fuga = energias_passadas_canal_fuga
        self.demanda_liquida_energia = demanda_liquida_energia
        self.convergencia = convergencia
        self.risco_ens = risco_ens
        self.custo_series_simuladas = custo_series_simuladas
        self.valor_esperado_periodo = valor_esperado_periodo
        self.custo_referenciado = custo_referenciado

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre PMO avalia todos os campos,
        menos a versão do NEWAVE e a convergência.
        """
        if not isinstance(o, PMO):
            return False
        pmo: PMO = o
        dif = False
        for (k, u), (_, v) in zip(self.__dict__.items(),
                                  pmo.__dict__.items()):
            if k == "versao_newave" or k == "convergencia":
                continue
            if u != v:
                dif = True
                break
        return not dif



# Rotinas de testes associadas ao arquivo cmarg00x.out do NWLISTOP
from inewave.newave.patamar import LeituraPatamar
from inewave.nwlistop.cmarg00 import LeituraCmarg00
from inewave.config import MESES, NUM_CENARIOS
import numpy as np  # type: ignore


sub_teste = "SUDESTE"
leitor = LeituraCmarg00("tests/_arquivos")
leitor.le_arquivos()


def test_leitura():
    assert sub_teste in leitor.cmargs


def test_eq_cmarg00():
    leitor2 = LeituraCmarg00("tests/_arquivos")
    leitor2.le_arquivos()
    assert leitor.cmargs == leitor2.cmargs


def test_neq_cmarg00():
    leitor2 = LeituraCmarg00("tests/_arquivos")
    leitor2.le_arquivos()
    leitor2.cmargs[sub_teste].ano_pmo = 2077
    assert leitor.cmargs != leitor2.cmargs


def test_extrai_dados_execucao():
    cmarg = leitor.cmargs[sub_teste]
    assert cmarg.mes_pmo == 5
    assert cmarg.ano_pmo == 1995
    assert cmarg.submercado == sub_teste


def test_anos_estudo():
    cmarg = leitor.cmargs[sub_teste]
    anos_estudo_teste = [1995, 1996, 1997, 1998, 1999]
    anos_lidos = list(cmarg.custos_patamares.keys())
    assert anos_estudo_teste == anos_lidos


def test_cmarg_por_patamar():
    cmarg = leitor.cmargs[sub_teste]
    por_patamar = cmarg.custos_por_patamar
    assert [1, 2, 3] == list(por_patamar.keys())
    mes = cmarg.mes_pmo
    ano = cmarg.ano_pmo
    # Confere se os valores médios de custos nos meses anteriores
    # ao estudo são nulos, e todos os posteriores são > 0
    for m in range(1, mes):
        assert np.mean(por_patamar[1][ano][m]) == 0.0
    for m in range(mes, len(MESES) + 1):
        assert np.mean(por_patamar[1][ano][m]) > 0.0
    for a in list(por_patamar[1].keys())[1:]:
        for m in range(mes, len(MESES) + 1):
            assert np.mean(por_patamar[1][a][m]) > 0.0


def test_cmarg_por_ano():
    cmarg = leitor.cmargs[sub_teste]
    por_ano = cmarg.custos_por_ano
    assert [1, 2, 3] == list(por_ano.keys())
    mes = cmarg.mes_pmo
    ano = cmarg.ano_pmo
    # Confere se os valores médios de custos nos meses anteriores
    # ao estudo são nulos, e todos os posteriores são > 0
    assert np.mean(por_ano[1][ano][:, :mes-1]) == 0.0
    assert np.mean(por_ano[1][ano][:, mes-1:]) > 0.0
    for a in list(por_ano[1].keys())[1:]:
        assert np.mean(por_ano[1][a]) > 0.0


def test_cmarg_por_ano_e_mes():
    cmarg = leitor.cmargs[sub_teste]
    por_ano_e_mes = cmarg.custos_por_ano_e_mes
    assert [1, 2, 3] == list(por_ano_e_mes.keys())
    mes = cmarg.mes_pmo
    ano = cmarg.ano_pmo
    # Confere se os valores médios de custos nos meses anteriores
    # ao estudo são nulos, e todos os posteriores são > 0
    for m in range(1, mes):
        assert np.mean(por_ano_e_mes[1][ano][m]) == 0.0
    for m in range(mes, len(MESES) + 1):
        assert np.mean(por_ano_e_mes[1][ano][m]) > 0.0
    for a in list(por_ano_e_mes[1].keys())[1:]:
        for m in range(mes, len(MESES) + 1):
            assert np.mean(por_ano_e_mes[1][a][m]) > 0.0


def test_cmarg_medio_por_ano():
    cmarg = leitor.cmargs[sub_teste]
    le_patamar = LeituraPatamar("tests/_arquivos")
    patamar = le_patamar.le_arquivo()
    medios = cmarg.custos_medios_por_ano(patamar)
    n_meses = len(MESES)
    mes_pmo = cmarg.mes_pmo
    # Confere se os valores médios segundo os patamares
    # são nulos até o mês anterior ao PMO
    anos = patamar.anos_estudo
    assert np.all(medios[anos[0]][:, :mes_pmo-1] == 0)
    for a in anos:
        assert medios[a].shape == (NUM_CENARIOS, n_meses)


def test_cmarg_medio_por_ano_e_mes():
    cmarg = leitor.cmargs[sub_teste]
    le_patamar = LeituraPatamar("tests/_arquivos")
    patamar = le_patamar.le_arquivo()
    medios = cmarg.custos_medios_por_ano_e_mes(patamar)
    n_meses = len(MESES)
    mes_pmo = cmarg.mes_pmo
    # Confere se os valores médios segundo os patamares
    # são nulos até o mês anterior ao PMO
    anos = patamar.anos_estudo
    for m in range(1, mes_pmo):
        assert np.all(medios[anos[0]][m] == 0)
    for a in anos:
        for m in range(1, n_meses + 1):
            custos: np.ndarray = medios[a][m]
            assert custos.shape == (NUM_CENARIOS,)



# Imports do próprio módulo
from inewave.config import REES
from inewave._utils.leitura import Leitura
from .modelos.estados import Estados, RegistroEstado
# Imports de módulos externos
import os
import numpy as np  # type: ignore
from traceback import print_exc
from typing import IO, Dict


class LeituraEstados(Leitura):
    """
    Realiza a leitura do arquivo estados.rel,
    existente em um diretório de saídas do NWLISTCF.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos do arquivo estados.rel, construindo um
    objeto `Estados` cujas informações são as mesmas do arquivo.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.

    Uma vez realizada a leitura do arquivo, as informações são guardadas
    internamente no atributo `estados`.

    **Exemplos**

    >>> diretorio = "~/documentos/.../deck"
    >>> leitor = LeituraEstados(diretorio)
    >>> leitor.le_arquivo()
    # Ops, esqueci de pegar o objeto
    >>> estados = leitor.estados

    """
    str_inicio_periodo = "  PERIODO:      "

    def __init__(self,
                 diretorio: str) -> None:
        super().__init__()
        self.diretorio = diretorio
        self.estados = Estados({})

    def le_arquivo(self) -> Estados:
        """
        Faz a leitura do arquivo `estados.rel`.
        """
        try:
            caminho = os.path.join(self.diretorio, "estados.rel")
            with open(caminho, "r") as arq:
                while True:
                    linha = self._le_linha_com_backup(arq)
                    # Verifica se o arquivo acabou
                    if len(linha) < 2:
                        break
                    # Verifica se a próxima linha indica um novo período
                    if LeituraEstados.str_inicio_periodo in linha:
                        periodo = int(linha[12:].strip())
                        p = self._le_periodo(arq)
                        self.estados.registros[periodo] = p
                return self.estados
        except Exception:
            print_exc()
            return self.estados

    def _le_periodo(self, arq: IO) -> Dict[int, RegistroEstado]:
        """
        Lê um período do arquivo estados.rel.
        """
        # Salta duas linhas para acessar a tabela
        self._le_linha_com_backup(arq)
        self._le_linha_com_backup(arq)
        p: Dict[int, RegistroEstado] = {}
        while True:
            # Verifica se a próxima linha é o início do
            # próximo período ou é vazia
            linha = self._le_linha_com_backup(arq)
            if (LeituraEstados.str_inicio_periodo in linha
                    or len(linha) < 2):
                break
            # Senão, lê mais um registro
            self._configura_backup()
            reg = self._le_registro(arq)
            p[reg.ireg] = reg

        return p

    def _le_registro(self, arq: IO) -> RegistroEstado:
        """
        """
        primeira_linha = True
        n_rees = len(REES)
        n_cols_tabela = 18
        ireg = 0
        itec = 0
        simc = 0
        itef = 0
        fobj = 0.
        tabela = np.zeros((n_rees, n_cols_tabela))
        for i in range(n_rees):
            linha = self._le_linha_com_backup(arq)
            if primeira_linha:
                primeira_linha = False
                # Extrai os campos específicos da primeira linha
                ireg = int(linha[2:10])
                itec = int(linha[11:15])
                simc = int(linha[16:20])
                itef = int(linha[21:25])
                fobj = float(linha[31:48])
            # Preenche a tabela com os dados do registro
            ree = int(linha[26:30])
            ci = 49
            nc = 17
            for j in range(n_cols_tabela):
                cf = ci + nc
                num_str = linha[ci:cf]
                valor = 0.0 if not num_str.isnumeric() else float(num_str)
                tabela[ree-1, j] = valor
                ci = cf + 1

        return RegistroEstado(ireg,
                              itec,
                              simc,
                              itef,
                              fobj,
                              tabela)

    def _fim_arquivo(self, linha: str) -> bool:
        return super()._fim_arquivo(linha)
